# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from numpy import array, sum, sqrt, log, abs, loadtxt, zeros_like, shape
from matplotlib.pyplot import plot, figure, subplot, show, ion, title, suptitle, legend, gca, ioff, axvline, gcf
import subprocess, sys, os, time
from six.moves import range
from builtins import str

print()
print('  ########  start script rms_tracker  ########')
print('  please give input path and some options:')
print('  available options are: 0 = rms')
print('                         1 = total energy')
print('                         2 = charge neutrality')
print('                         3 = total moment')
print()
print('  if last given option is -1 then no refreshing is done and plots are only generated once')
print()
print(
    "  after the options there is the possibility to give the 'nodos' option which prevents the plotting of the dos if 'dos.atom*' files are there"
)
print()
print('  input might look like this:')
print('    for self refreshing plot of rms charge neutrality and moment: ./rms_tracker.py ./path/ 0 2 3')
print('    for non refreshing plot of rms and total energy without dos plot: ./rms_tracker.py ./path/ 0 1 -1 nodos')
print()
print('  default values: ./rms_tracker.py ./path/ 0 1 2 3')

if len(sys.argv) > 1:
    allpaths = sys.argv[1]
    # create list of paths from input
    allpath = []
    for path0 in allpaths.split(','):
        outfile_found = False
        for file in ['out', 'out_kkr']:
            if file in os.listdir(path0):
                outfile_found = True
                outfile = file
        if outfile_found:
            if path0 == '.' or path0 == './':
                path0 = os.getcwd()
            if path0[-1] != '/':
                path0 += '/'
            allpath.append(path0)
        else:
            print("WARNING: file 'out' not found in", path0)
    if allpath == []:
        print("no file 'out' found")
        sys.exit()

    if len(sys.argv) > 2:
        nopt = len(sys.argv[2:])
        opt = sys.argv[2:]
        if len(opt[-1]) <= 2:
            dos = 'dos.atom1' in os.listdir(path0)
        else:
            dos = False
            nopt = nopt - 1
            opt = opt[:-1]
        opt = [int(i) for i in opt]
        print(opt)
        if opt[-1] == -1:
            refresh = False
            nopt = nopt - 1
            opt = opt[:-1]
        else:
            refresh = True
    else:
        nopt = 4
        opt = [0, 1, 2, 3]
        refresh = True
        if len(sys.argv[-1]) <= 2:
            dos = 'dos.atom1' in os.listdir(path0)
        else:
            dos = False

else:
    print('Please give input path')
    sys.exit()

# consistency check and names
names = []
for i in opt:
    if i not in [0, 1, 2, 3]:
        print('Found non existing option:', i)
        sys.exit()
    else:
        if i == 0:
            names.append('rms error')
        elif i == 1:
            names.append('total energy')
        elif i == 2:
            names.append('neutrality')
        elif i == 3:
            names.append('total moment')

# print input
print()
print('  your input options:', opt, '; refresh=', refresh, '; DOS=', dos)
print('  path(s):', allpath)
print()


def read_rms_data(path0):
    global opt, nopt
    ### rms
    try:
        f = subprocess.check_output('grep aver ' + path0 + outfile, shell=True).decode('utf-8').split('\n')
        rms = array([float((i.split()[-1]).replace('D', 'e').replace('\n', '')) for i in f if i != ''])
    except:
        rms = []

    ### charge neutrality
    try:
        f = subprocess.check_output('grep neutr ' + path0 + outfile, shell=True).decode('utf-8').split('\n')
        neut = array([float((i.split()[-1]).replace('D', 'e').replace('\n', '')) for i in f if i != ''])
    except:
        neut = []
    # check if neutrality info is available (not there in case of impurity code output)
    if len(neut) < len(rms) and 2 in opt:
        nopt = nopt - 1
        j = 0
        for i in range(nopt + 1):
            if opt[i] == 2:
                j = i
        print('removing option', opt.pop(j), names.pop(j))

    ### read total energy
    try:
        f = subprocess.check_output("grep 'TOTAL ENERGY' " + path0 + outfile, shell=True).decode('utf-8').split('\n')
        etot = array([float((i.split()[-1]).replace('D', 'e').replace('\n', '')) for i in f if i != ''])
    except:
        etot = []

    ### moment
    try:
        f = subprocess.check_output("grep 'L m' " + path0 + outfile, shell=True).decode('utf-8').split('\n')
        mom = array([float((i.split()[-1]).replace('D', 'e').replace('\n', '')) for i in f if i != ''])
    except:
        mom = []
    # check if moment info is available
    if len(mom) < len(rms) and 3 in opt:
        nopt = nopt - 1
        for i in range(nopt + 1):
            if opt[i] == 3:
                j = i
        print('removing option', opt.pop(j), names.pop(j))
    else:
        try:
            f = subprocess.check_output("grep 'ITERATION ' " + path0 + outfile, shell=True).decode('utf-8').split('\n')
            tmp = array(
                [float((i.split()[1]).replace('D', 'e').replace('\n', '')) for i in f if 'SCF' not in i and i != ''])
        except:
            tmp = []
        if tmp != [] and 'out_magneticmoments' in os.listdir(path0):
            it = int(tmp[-1])
            tmp = loadtxt(path0 + 'out_magneticmoments')
            ncls = len(tmp) // it
            tmp = tmp[::ncls, :3]
            tmpmomx = tmp[:, 0]
            tmpmomy = tmp[:, 1]
            tmpmomz = tmp[:, 2]
            tmpmom = mom
            mom = array([tmpmom, tmpmomx, tmpmomy, tmpmomz])
        else:
            inp = open(path0 + 'inputcard').readlines()
            natyp = -1
            for iline in range(len(inp)):
                if 'NATYP' in inp[iline]:
                    natyp = iline
            if natyp == -1:
                for iline in range(len(inp)):
                    if 'NAEZ' in inp[iline]:
                        natyp = iline
                natyp = int(inp[natyp].split('NAEZ=')[1])
            else:
                natyp = int(inp[natyp].split('NATYP=')[1])
            try:
                tmp2 = subprocess.check_output("grep 'm_spin' " + path0 + outfile + ' -A' + str(natyp),
                                               shell=True).decode('utf-8').replace('TOT', '').split('\n')
                if tmp2 != '':
                    tmp2 = tmp2  #[-natyp-1:]
                    tmp2 = array([float(i.split()[2]) for i in tmp2 if i != '' and 'dn' not in i and i != '--'])
                    tmp2 = tmp2.reshape(-1, natyp)
                    mom = array([mom] + [tmp2[:, i] for i in range(len(tmp2[0]))])
            except:
                tmpmom, tmpmomx, tmpmomy, tmpmomz = mom, mom, mom, mom
                mom = array([tmpmom, tmpmomx, tmpmomy, tmpmomz])

    return rms, neut, mom, etot


# turn interactive mode on (needed for figures to appear immediately in while loop)
ion()

rms = [[] for i in range(len(allpath))]
neut = [[] for i in range(len(allpath))]
etot = [[] for i in range(len(allpath))]
mom = [[] for i in range(len(allpath))]
istart = [True for i in range(len(allpath))]

while (True):
    for ipath0 in range(len(allpath)):
        path0 = allpath[ipath0]
        l0 = len(rms[ipath0])
        rms[ipath0], neut[ipath0], mom[ipath0], etot[ipath0] = read_rms_data(path0)

        if len(rms[ipath0]) > l0:
            ### fig 1 for plots of values from option
            figure(ipath0)
            figure(ipath0).clf()

            # give path name to figure
            gcf().canvas.set_window_title(allpath[ipath0])

            for i in range(nopt):
                # choose data according to option
                if opt[i] == 0:
                    data = rms[ipath0]
                elif opt[i] == 1:
                    data = etot[ipath0]
                elif opt[i] == 2:
                    data = neut[ipath0]
                elif opt[i] == 3:
                    data = mom[ipath0]

                # choose appropriate subplot according to number of options
                if nopt == 1:
                    subplot(1, 2, 1 + i)
                elif nopt == 2:
                    subplot(2, 2, 1 + i)
                elif nopt == 3:
                    subplot(2, 3, 1 + i)
                elif nopt == 4:
                    subplot(2, 4, 1 + i)

                # plot data in linear scale
                title(names[i])
                if opt[i] == 3:
                    clrs = ['b', 'g', 'r', 'y', 'm', 'c', 'k']
                    for j in range(len(data) - 1, -1, -1):
                        plot(data[j], 'x--')  #+clrs[j])
                else:
                    plot(data, 'x--')

                # substract reference for Etot and mom
                if len(data) > 1 and opt[i] in [1, 3]:
                    if opt[i] == 3 and len(data[0]) > 1:
                        datanew = zeros_like(data)[:, :-1]
                        for j in range(len(data[:, 0])):
                            datanew[j, :] = (data[j, :] - data[j, -1])[:-1]
                        data = datanew
                    else:
                        data = (data - data[-1])[:-1]

                #  plot log scale if data is not always 0
                if (abs(data).any() > 0):
                    if nopt == 1:
                        subplot(1, 2, 1 + i + nopt)
                    elif nopt == 2:
                        subplot(2, 2, 1 + i + nopt)
                    elif nopt == 3:
                        subplot(2, 3, 1 + i + nopt)
                    elif nopt == 4:
                        subplot(2, 4, 1 + i + nopt)
                    if opt[i] in [1, 3]:
                        name = 'change in ' + names[i]
                    else:
                        name = names[i] + ', log scale'
                    title(name)
                    if opt[i] == 3:
                        for j in range(len(data) - 1, -1, -1):
                            plot(abs(data[j]), 'x--')  #+clrs[j])
                    else:
                        plot(abs(data), 'x--')
                    gca().set_yscale('log')
                try:
                    t_iter = subprocess.check_output('grep Iter ' + path0 + 'out_timing.000.txt',
                                                     shell=True).decode('utf-8').split('\n')[-2].split()[-1]
                    t_iter = '%5.2f min' % (float(t_iter) / 60)
                except:
                    t_iter = 'no time info for last iteration'
                suptitle(time.ctime() + ', time in last iteration: ' + t_iter)

            ### fig 2 for DOS plots
            if dos:
                figure(ipath0 + 1000)
                figure(ipath0 + 1000).clf()
                j = 0
                for i in os.listdir(path0):
                    if len(i) > 8 and 'dos.atom' == i[:8]:
                        tmpi = int(i.replace('dos.atom', ''))
                        if j <= tmpi:
                            j = tmpi

                if len(rms[ipath0]) > 1 and not istart[ipath0]:
                    subplot(1, 2, 2)
                    title('previous iteration')
                    for i in range(len(d_old)):
                        d = d_old[i]
                        lab = i
                        plot(d[:, 0], sum(d[:, 1:], axis=1), '-', label=lab)
                    legend()
                    ef = float(open(path0 + 'potential').readlines()[3].split()[1])
                    axvline(ef)
                    subplot(1, 2, 1)
                else:
                    subplot(1, 1, 1)
                title('current iteration')
                d_old = []
                for i in range(j):
                    d = loadtxt((path0 + 'dos.atom%i' % (i + 1)))
                    lab = i
                    plot(d[:, 0], sum(d[:, 1:], axis=1), '-', label=lab)
                    d_old.append(d)
                    istart[ipath0] = False
                legend()
                ef = float(open(path0 + 'potential').readlines()[3].split()[1])
                axvline(ef)
                suptitle(time.ctime())

        figure(ipath0).canvas.draw()
        figure(ipath0).canvas.flush_events()

        if dos:
            figure(ipath0 + 1000).canvas.draw()
            figure(ipath0 + 1000).canvas.flush_events()

        dobreak = False
        if refresh:
            time.sleep(2)
        else:
            # turn off interactive mode and exit while loop
            if path0 == allpath[-1]:
                dobreak = True
    if dobreak:
        break

ioff()
show()



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), 2018 Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.    #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools    #
# For further information on the license, see the LICENSE.txt file            #
#                                                                             #
###############################################################################
"""
This module defines useful utility for logging related functionality
"""
from logging import Handler, LoggerAdapter


class DictHandler(Handler):
    """
    Custom Handler for the logging module inserting logging messages
    into a given dictionary.

    Messages are grouped into list under the names of the error categories.
    Keyword arguments can be used to modify the keys for the different levels
    """

    def __init__(self, log_dict, ignore_unknown_levels=False, **kwargs):
        from logging import _levelToName
        import copy

        self.log_dict = log_dict

        levels = copy.copy(list(_levelToName.values()))
        levels.remove('NOTSET')

        self.level_names = {name: kwargs[name] for name in levels if name in kwargs}

        if not ignore_unknown_levels:
            for name in levels:
                if name not in self.level_names:
                    self.level_names[name] = name

        for name in self.level_names.values():
            self.log_dict[name] = []

        super().__init__()

    def emit(self, record):
        """
        Emit a record.
        """
        try:
            msg = self.format(record)
            entry_name = self.level_names.get(record.levelname, 'NOTSET')
            if entry_name not in self.log_dict:
                self.log_dict[entry_name] = []

            self.log_dict[entry_name].append(msg)
        except Exception:  #pylint: disable=broad-except
            self.handleError(record)

    def __repr__(self):
        from logging import getLevelName
        level = getLevelName(self.level)
        return '<%s (%s)>' % (self.__class__.__name__, level)


class OutParserLogAdapter(LoggerAdapter):
    """
    This adapter expects the passed in dict-like object to have a
    'iteration' key, whose value is prepended as [Iteration i] to the message
    """

    def process(self, msg, kwargs):
        return '[Iteration %s] %s' % (self.extra['iteration'], msg), kwargs



# -*- coding: utf-8 -*-
"""
Test of the functions in masci_tools.util.xml.converters
"""
import pytest
import os
from masci_tools.util.constants import FLEUR_DEFINED_CONSTANTS
from pprint import pprint
import logging

TEST_FOLDER = os.path.dirname(os.path.abspath(__file__))
CLEAR_XML_TEST_FILE = os.path.abspath(os.path.join(TEST_FOLDER, 'files/fleur/test_clear.xml'))

LOGGER = logging.getLogger(__name__)


def test_convert_from_fortran_bool():
    """
    Test of the convert_from_fortran_bool function
    """
    from masci_tools.util.xml.converters import convert_from_fortran_bool

    TRUE_ITEMS = ('T', 't', True)
    FALSE_ITEMS = ('F', 'f', False)

    for true_item in TRUE_ITEMS:
        assert convert_from_fortran_bool(true_item)

    for false_item in FALSE_ITEMS:
        assert not convert_from_fortran_bool(false_item)

    with pytest.raises(ValueError, match="Could not convert: 'TEST' to boolean"):
        convert_from_fortran_bool('TEST')

    with pytest.raises(TypeError, match="Could not convert: '{}' to boolean"):
        convert_from_fortran_bool({})


def test_convert_to_fortran_bool():
    """
    Test of the function convert_to_fortran_bool
    """
    from masci_tools.util.xml.converters import convert_to_fortran_bool

    TRUE_ITEMS = (True, 'True', 't', 'T')
    FALSE_ITEMS = (False, 'False', 'f', 'F')

    for item in TRUE_ITEMS:
        assert convert_to_fortran_bool(item) == 'T'

    for item in FALSE_ITEMS:
        assert convert_to_fortran_bool(item) == 'F'

    with pytest.raises(ValueError, match='A string: NOT_A_BOOL for a boolean was given,'):
        convert_to_fortran_bool('NOT_A_BOOL')

    with pytest.raises(TypeError, match='convert_to_fortran_bool accepts only a string or bool as argument'):
        convert_to_fortran_bool(())


TEST_STRINGS = ['1.2314', 'all', ['all', '213', '-12'], ['PI', 'NOT_PI', '1.2'], ['F', 'T'], ['F', 'T']]

TEST_TYPES = [['float'], ['int', 'string'], ['int', 'string'], ['float', 'float_expression'], ['int'],
              ['int', 'switch']]

TEST_RESULTS = [(pytest.approx(1.2314), True), ('all', True), (['all', 213, -12], True),
                (['PI', 'NOT_PI', pytest.approx(1.2)], False), (['F', 'T'], False), ([False, True], True)]

TEST_WARNINGS = [[], [], [], [
    "Could not convert 'PI'",
    "Could not convert 'NOT_PI'",
], ["Could not convert 'F'", "Could not convert 'T'"]]


@pytest.mark.parametrize('string_attr,types,results', zip(TEST_STRINGS, TEST_TYPES, TEST_RESULTS))
def test_convert_xml_attribute(string_attr, types, results):
    """
    Test of the convert_xml_attribute function
    """
    from masci_tools.util.xml.converters import convert_xml_attribute

    expected_val, expected_suc = results

    if not expected_suc:
        with pytest.raises(ValueError, match='Could not convert'):
            ret_val, suc = convert_xml_attribute(string_attr, types, FLEUR_DEFINED_CONSTANTS)
    else:
        ret_val, suc = convert_xml_attribute(string_attr, types, FLEUR_DEFINED_CONSTANTS)

        assert ret_val == expected_val
        assert suc == expected_suc


@pytest.mark.parametrize('string_attr,types,results, warnings',
                         zip(TEST_STRINGS, TEST_TYPES, TEST_RESULTS, TEST_WARNINGS))
def test_convert_xml_attribute_warnings(caplog, string_attr, types, results, warnings):
    """
    Test of the convert_xml_attribute function
    """
    from masci_tools.util.xml.converters import convert_xml_attribute

    expected_val, expected_suc = results

    with caplog.at_level(logging.WARNING):
        ret_val, suc = convert_xml_attribute(string_attr, types, FLEUR_DEFINED_CONSTANTS, logger=LOGGER)
    assert ret_val == expected_val
    assert suc == expected_suc

    if len(warnings) == 0:
        assert caplog.text == ''
    else:
        for expected_warning in warnings:
            assert expected_warning in caplog.text


TEST_ATTR_VALUES = [1.2134, 'all', ['all', 213, '-12'], ['3.14', 'NOT_PI', 1.2], [False, 'True']]

TEST_ATTR_TYPES = [['float'], ['int', 'string'], ['int', 'string'], ['float', 'float_expression'], ['int', 'switch']]

TEST_ATTR_RESULTS = [('1.2134000000', True), ('all', True), (['all', '213', '-12'], True),
                     (['3.14', 'NOT_PI', '1.2000000000'], True), (['F', 'T'], True)]

TEST_ATTR_WARNINGS = [[], [], [], [], []]


@pytest.mark.parametrize('attr_value,types,results', zip(TEST_ATTR_VALUES, TEST_ATTR_TYPES, TEST_ATTR_RESULTS))
def test_convert_attribute_to_xml(attr_value, types, results):
    """
    Test of the convert_xml_attribute function
    """
    from masci_tools.util.xml.converters import convert_attribute_to_xml

    expected_val, expected_suc = results
    if not expected_suc:
        with pytest.raises(ValueError):
            ret_val, suc = convert_attribute_to_xml(attr_value, types)
    else:
        ret_val, suc = convert_attribute_to_xml(attr_value, types)
        assert ret_val == expected_val
        assert suc == expected_suc


@pytest.mark.parametrize('attr_value,types,results,warnings',
                         zip(TEST_ATTR_VALUES, TEST_ATTR_TYPES, TEST_ATTR_RESULTS, TEST_ATTR_WARNINGS))
def test_convert_attribute_to_xml_warnings(caplog, attr_value, types, results, warnings):
    """
    Test of the convert_xml_attribute function
    """
    from masci_tools.util.xml.converters import convert_attribute_to_xml

    expected_val, expected_suc = results

    with caplog.at_level(logging.WARNING):
        ret_val, suc = convert_attribute_to_xml(attr_value, types, logger=LOGGER)
    assert ret_val == expected_val
    assert suc == expected_suc

    if len(warnings) == 0:
        assert caplog.text == ''
    else:
        for expected_warning in warnings:
            assert expected_warning in caplog.text


TEST_TEXT_STRINGS = [
    '0.0 Pi/4.0 6.3121', '0.0 Pi/4.0 6.3121', '0.0 Pi/4.0 6.3121', '0.0 Pi/4.0 6.3121',
    ['0.0 Pi/4.0 6.3121', 'Bohr Pi/4.0 all', '0.0 Pi/*4.0 0.0'], ['F asd', 'T'],
    ['12 213 4215 412', '12 213 4215 412 123 14124']
]

TEST_DEFINITIONS = [[{
    'length': 3,
    'type': ['float', 'float_expression']
}], [{
    'length': 4,
    'type': ['float', 'float_expression']
}],
                    [{
                        'length': 4,
                        'type': ['float', 'float_expression']
                    }, {
                        'length': 'unbounded',
                        'type': ['float', 'float_expression']
                    }], [{
                        'length': 1,
                        'type': ['float', 'float_expression']
                    }], [{
                        'length': 3,
                        'type': ['float', 'float_expression']
                    }], [{
                        'length': 'unbounded',
                        'type': ['switch', 'string']
                    }], [{
                        'length': 4,
                        'type': ['switch', 'int']
                    }, {
                        'length': 'unbounded',
                        'type': ['int']
                    }]]

TEST_TEXT_RESULTS = [(pytest.approx([0.0, 0.7853981633974483, 6.3121]), True), ('0.0 Pi/4.0 6.3121', False),
                     (pytest.approx([0.0, 0.7853981633974483, 6.3121]), True), ('0.0 Pi/4.0 6.3121', False),
                     ([[0.0, 0.7853981633974483, 6.3121], [1.0, 0.7853981633974483, 'all'], [0.0, 'Pi/*4.0',
                                                                                             0.0]], False),
                     ([[False, 'asd'], [True]], True), ([[12, 213, 4215, 412], [12, 213, 4215, 412, 123, 14124]], True)]

TEST_TEXT_WARNINGS = [[], ["Failed to convert '0.0 Pi/4.0 6.3121', no matching definition found"], [],
                      [
                          "Could not convert '0.0 Pi/4.0 6.3121'",
                      ], [
                          "Could not convert 'all'",
                          "Could not convert 'Pi/*4.0'",
                      ], [], []]


@pytest.mark.parametrize('string_text,definitions, results', zip(TEST_TEXT_STRINGS, TEST_DEFINITIONS,
                                                                 TEST_TEXT_RESULTS))
def test_convert_xml_text(string_text, definitions, results):
    """
    Test of the convert_xml_attribute function
    """
    from masci_tools.util.xml.converters import convert_xml_text

    expected_val, expected_suc = results

    if not expected_suc:
        with pytest.raises(ValueError):
            ret_val, suc = convert_xml_text(string_text, definitions, FLEUR_DEFINED_CONSTANTS)
    else:
        ret_val, suc = convert_xml_text(string_text, definitions, FLEUR_DEFINED_CONSTANTS)
        assert ret_val == expected_val
        assert suc == expected_suc


@pytest.mark.parametrize('string_text,definitions,results,warnings',
                         zip(TEST_TEXT_STRINGS, TEST_DEFINITIONS, TEST_TEXT_RESULTS, TEST_TEXT_WARNINGS))
def test_convert_xml_text_warnings(caplog, string_text, definitions, results, warnings):
    """
    Test of the convert_xml_attribute function
    """
    from masci_tools.util.xml.converters import convert_xml_text

    expected_val, expected_suc = results

    with caplog.at_level(logging.WARNING):
        ret_val, suc = convert_xml_text(string_text, definitions, FLEUR_DEFINED_CONSTANTS, logger=LOGGER)
    assert ret_val == expected_val
    assert suc == expected_suc
    if len(warnings) == 0:
        assert caplog.text == ''
    else:
        for expected_warning in warnings:
            assert expected_warning in caplog.text


TEST_TEXT_VALUES = [[0.0, 'Pi/4.0', 6.3121], [0.0, 'Pi/4.0', 6.3121], [0.0, 'Pi/4.0', 6.3121], [0.0, 'Pi/4.0', 6.3121],
                    [[0.0, 'Pi/4.0', 6.3121], ['Bohr', 'Pi/4.0', 'all']], [[False, 'asd'], 'T'],
                    [[12, '213', 4215, 412], [12, '213', 4215, '412', 123, 14124]]]

TEST_TEXT_XML_STRINGS = [(' 0.0000000000000 Pi/4.0  6.3121000000000', True), ('', False),
                         (' 0.0000000000000 Pi/4.0  6.3121000000000', True), ('', False),
                         ([' 0.0000000000000 Pi/4.0  6.3121000000000', 'Bohr Pi/4.0 all'], True), (['F asd',
                                                                                                    'T'], True),
                         (['12 213 4215 412', '12 213 4215 412 123 14124'], True)]

TEST_TEXT_XML_WARNINGS = [[], ["Failed to convert '[0.0, 'Pi/4.0', 6.3121]', no matching definition found"], [],
                          ["Failed to convert '[0.0, 'Pi/4.0', 6.3121]', no matching definition found"], [], [], []]


@pytest.mark.parametrize('text_value,definitions, results',
                         zip(TEST_TEXT_VALUES, TEST_DEFINITIONS, TEST_TEXT_XML_STRINGS))
def test_convert_text_to_xml(text_value, definitions, results):
    """
    Test of the convert_xml_attribute function
    """
    from masci_tools.util.xml.converters import convert_text_to_xml

    expected_val, expected_suc = results
    if not expected_suc:
        with pytest.raises(ValueError):
            ret_val, suc = convert_text_to_xml(text_value, definitions)
    else:
        ret_val, suc = convert_text_to_xml(text_value, definitions)
        assert ret_val == expected_val
        assert suc == expected_suc


@pytest.mark.parametrize('text_value,definitions, results, warnings',
                         zip(TEST_TEXT_VALUES, TEST_DEFINITIONS, TEST_TEXT_XML_STRINGS, TEST_TEXT_XML_WARNINGS))
def test_convert_text_to_xml_warnings(caplog, text_value, definitions, results, warnings):
    """
    Test of the convert_xml_attribute function
    """
    from masci_tools.util.xml.converters import convert_text_to_xml

    expected_val, expected_suc = results

    with caplog.at_level(logging.WARNING):
        ret_val, suc = convert_text_to_xml(text_value, definitions, logger=LOGGER)
    assert ret_val == expected_val
    assert suc == expected_suc
    if len(warnings) == 0:
        assert caplog.text == ''
    else:
        for expected_warning in warnings:
            assert expected_warning in caplog.text



# -*- coding: utf-8 -*-
"""
Collect all functions that should be exposed to the :py:class:`~masci_tools.io.io_fleurxmlmodifier.FleurXMLModifier`
and classify them according to their interface. This makes extending functionality for the Modifier relatively easy

   1. Implement the method in the correct module for its level of functionality
         :xml_setters_basic: Xpath as argument and no ability to create missing subtags on the fly
         :xml_setters_xpaths: Xpath(s) and schema dictionary as argument. Can add creation type/checking and so on
         :xml_setters_names: No Xpaths as arguments Only names/identifiers for the schema dict
   2. Import it in this file and put the method in the correct SET at the bottom
   3. Write facade method to add task for this function in :py:class:`~masci_tools.io.io_fleurxmlmodifier.FleurXMLModifier`
      This is not automatized to guarantee that we have some nice docstrings for all the modifiying functions at the surface level
"""

from .xml_setters_names import set_inpchanges, shift_value, set_species, set_species_label, shift_value_species_label,\
                               set_complex_tag, set_simple_tag, create_tag, set_text, set_first_text, set_attrib_value,\
                               set_first_attrib_value, set_atomgroup, set_atomgroup_label, add_number_to_attrib, add_number_to_first_attrib

from .xml_setters_basic import xml_create_tag, xml_set_attrib_value_no_create, xml_set_text_no_create, \
                               xml_replace_tag, xml_delete_tag, xml_delete_att

from .xml_setters_nmmpmat import set_nmmpmat, rotate_nmmpmat

__XPATH_SET = {
    xml_create_tag, xml_set_text_no_create, xml_set_attrib_value_no_create, xml_replace_tag, xml_delete_att,
    xml_delete_tag
}

__SCHEMA_DICT_SET = {
    set_inpchanges, shift_value, set_species, set_species_label, shift_value_species_label, set_complex_tag,
    set_simple_tag, create_tag, set_text, set_first_text, set_attrib_value, set_first_attrib_value, set_atomgroup,
    set_atomgroup_label, add_number_to_attrib, add_number_to_first_attrib
}

__NMMPMAT_SET = {set_nmmpmat, rotate_nmmpmat}

XPATH_SETTERS = {func.__name__: func for func in __XPATH_SET}
SCHEMA_DICT_SETTERS = {func.__name__: func for func in __SCHEMA_DICT_SET}
NMMPMAT_SETTERS = {func.__name__: func for func in __NMMPMAT_SET}



# -*- coding: utf-8 -*-
"""
Tests for the functions in xml_setters_names

These tests do not extensively test all possible functionality. this is done in the
tests for the underlying functions in xml_setters_xpaths and xml_setters_basic
"""
import os
from lxml import etree
import pytest

FILE_PATH = os.path.dirname(os.path.abspath(__file__))
TEST_INPXML_PATH = os.path.join(FILE_PATH, 'files/fleur/Max-R5/FePt_film_SSFT_LO/files/inp2.xml')


def test_create_tag(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import create_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    node = eval_xpath(root, '/fleurInput/calculationSetup')

    tags = [child.tag for child in node.iterchildren()]
    tags.append('greensFunction')

    create_tag(xmltree, schema_dict, 'greensFunction')

    node = eval_xpath(root, '/fleurInput/calculationSetup')

    assert [child.tag for child in node.iterchildren()] == tags


def test_create_tag_specification(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import create_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    with pytest.raises(ValueError, match='The tag lo has multiple possible paths with the current specification.'):
        create_tag(xmltree, schema_dict, 'lo')

    create_tag(xmltree, schema_dict, 'lo', contains='species')

    los_after = eval_xpath(root, '/fleurInput/atomSpecies/species/lo')

    assert [node.getparent().attrib['name'] for node in los_after] == ['Fe-1', 'Fe-1', 'Fe-1', 'Pt-1', 'Pt-1']
    assert [node.attrib.items() for node in los_after] == [[],
                                                           [('type', 'SCLO'), ('l', '0'), ('n', '3'), ('eDeriv', '0')],
                                                           [('type', 'SCLO'), ('l', '1'), ('n', '3'), ('eDeriv', '0')],
                                                           [],
                                                           [('type', 'SCLO'), ('l', '1'), ('n', '5'), ('eDeriv', '0')]]


def test_create_tag_create_parents(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import create_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    node = eval_xpath(root, '/fleurInput/calculationSetup')

    tags = [child.tag for child in node.iterdescendants()]
    tags.extend(['greensFunction', 'realAxis'])

    with pytest.raises(ValueError, match="Could not create tag 'realAxis' because atleast one subtag is missing."):
        create_tag(xmltree, schema_dict, 'realAxis')

    create_tag(xmltree, schema_dict, 'realAxis', create_parents=True)

    node = eval_xpath(root, '/fleurInput/calculationSetup')

    assert [child.tag for child in node.iterdescendants()] == tags


def test_create_tag_complex_xpath(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import create_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    create_tag(xmltree,
               schema_dict,
               'lo',
               contains='species',
               complex_xpath="/fleurInput/atomSpecies/species[@name='Fe-1']")

    los_after = eval_xpath(root, '/fleurInput/atomSpecies/species/lo')

    assert [node.getparent().attrib['name'] for node in los_after] == ['Fe-1', 'Fe-1', 'Fe-1', 'Pt-1']
    assert [node.attrib.items() for node in los_after] == [[],
                                                           [('type', 'SCLO'), ('l', '0'), ('n', '3'), ('eDeriv', '0')],
                                                           [('type', 'SCLO'), ('l', '1'), ('n', '3'), ('eDeriv', '0')],
                                                           [('type', 'SCLO'), ('l', '1'), ('n', '5'), ('eDeriv', '0')]]


def test_create_tag_occurrences(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import create_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    create_tag(xmltree, schema_dict, 'lo', contains='species', occurrences=[-1])

    los_after = eval_xpath(root, '/fleurInput/atomSpecies/species/lo')

    assert [node.getparent().attrib['name'] for node in los_after] == ['Fe-1', 'Fe-1', 'Pt-1', 'Pt-1']
    assert [node.attrib.items() for node in los_after] == [[('type', 'SCLO'), ('l', '0'), ('n', '3'), ('eDeriv', '0')],
                                                           [('type', 'SCLO'), ('l', '1'), ('n', '3'), ('eDeriv', '0')],
                                                           [],
                                                           [('type', 'SCLO'), ('l', '1'), ('n', '5'), ('eDeriv', '0')]]


def test_set_attrib_value(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_attrib_value

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    set_attrib_value(xmltree, schema_dict, 'kmax', 5.321)

    kmax = eval_xpath(root, '/fleurInput/calculationSetup/cutoffs/@Kmax')

    assert kmax == '5.3210000000'


def test_set_attrib_value_xcFunctional(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_attrib_value

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    set_attrib_value(xmltree, schema_dict, 'xcFunctional', 'TEST')

    res = eval_xpath(root, '/fleurInput/calculationSetup/xcFunctional/@name')

    assert res == 'TEST'


def test_set_attrib_value_specification(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_attrib_value

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    with pytest.raises(ValueError,
                       match='The attrib radius has multiple possible paths with the current specification.'):
        set_attrib_value(xmltree, schema_dict, 'radius', [40, 42])

    set_attrib_value(xmltree, schema_dict, 'radius', [40, 42], contains='species')

    radius = eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius')

    assert radius == ['40.0000000000', '42.0000000000']


def test_set_attrib_value_create(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_attrib_value

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    with pytest.raises(
            ValueError,
            match=
            "Could not set attribute 'ne' on path '/fleurInput/calculationSetup/greensFunction/realAxis' because atleast one subtag is missing."
    ):
        set_attrib_value(xmltree, schema_dict, 'ne', 1000, contains='realAxis')

    set_attrib_value(xmltree, schema_dict, 'ne', 1000, contains='realAxis', create=True)

    ne = eval_xpath(root, '/fleurInput/calculationSetup/greensFunction/realAxis/@ne')

    assert ne == '1000'


def test_set_attrib_value_complex_xpath(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_attrib_value

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    set_attrib_value(xmltree,
                     schema_dict,
                     'type',
                     'TEST',
                     contains='species',
                     complex_xpath="/fleurInput/atomSpecies/species[@name='Fe-1']/lo")

    lo_types = eval_xpath(root, '/fleurInput/atomSpecies/species/lo/@type')

    assert lo_types == ['TEST', 'TEST', 'SCLO']


def test_set_attrib_value_occurrences(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_attrib_value

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    set_attrib_value(xmltree, schema_dict, 'type', 'TEST', contains='species', occurrences=-2)

    lo_types = eval_xpath(root, '/fleurInput/atomSpecies/species/lo/@type')

    assert lo_types == ['SCLO', 'TEST', 'SCLO']


def test_set_first_attrib_value(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_first_attrib_value

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    set_first_attrib_value(xmltree, schema_dict, 'type', 'TEST', contains='species')

    lo_types = eval_xpath(root, '/fleurInput/atomSpecies/species/lo/@type')

    assert lo_types == ['TEST', 'SCLO', 'SCLO']


def test_set_first_attrib_value_complex_xpath(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_first_attrib_value

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    set_first_attrib_value(xmltree,
                           schema_dict,
                           'type',
                           'TEST',
                           contains='species',
                           complex_xpath="/fleurInput/atomSpecies/species[@name='Pt-1']/lo")

    lo_types = eval_xpath(root, '/fleurInput/atomSpecies/species/lo/@type')

    assert lo_types == ['SCLO', 'SCLO', 'TEST']


def test_set_first_attrib_value_create(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_first_attrib_value

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    with pytest.raises(
            ValueError,
            match=
            "Could not set attribute 'U' on path '/fleurInput/atomSpecies/species/ldaU' because atleast one subtag is missing."
    ):
        set_first_attrib_value(xmltree, schema_dict, 'U', 42, contains={'species', 'ldaU'})

    set_first_attrib_value(xmltree, schema_dict, 'U', 42, contains={'species', 'ldaU'}, create=True)

    ldaU_node = eval_xpath(root, '/fleurInput/atomSpecies/species/ldaU')

    assert ldaU_node.attrib == {'U': '42.0000000000'}


def test_set_text(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_text

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    set_text(xmltree, schema_dict, 'comment', 'This is a test comment')

    res = eval_xpath(root, '/fleurInput/comment/text()')

    assert res == 'This is a test comment'


def test_set_text_specification_create(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_text

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    with pytest.raises(ValueError, match='The tag s has multiple possible paths with the current specification'):
        set_text(xmltree, schema_dict, 's', [False, False, False, True])

    with pytest.raises(
            ValueError,
            match=
            "Could not set text on path '/fleurInput/atomSpecies/species/torgueCalculation/greensfElements/s' because atleast one subtag is missing."
    ):
        set_text(xmltree, schema_dict, 's', [False, False, False, True], contains={'species', 'torgue'})

    set_text(xmltree, schema_dict, 's', [False, False, False, True], contains={'species', 'torgue'}, create=True)

    res = eval_xpath(root, '/fleurInput/atomSpecies/species/torgueCalculation/greensfElements/s/text()')

    assert res == ['F F F T', 'F F F T']


def test_set_text_specification_complex_xpath(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_text

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    set_text(xmltree,
             schema_dict,
             'kPoint', [10.0, 10.0, 10.0],
             complex_xpath='/fleurInput/cell/bzIntegration/kPointLists/kPointList/kPoint[1]')

    res = eval_xpath(root, '/fleurInput/cell/bzIntegration/kPointLists/kPointList/kPoint/text()')

    assert res == ['10.0000000000000 10.0000000000000 10.0000000000000', '    0.250000     0.250000     0.000000']


def test_set_text_specification_occurrences(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_text

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    set_text(xmltree, schema_dict, 'kPoint', [10.0, 10.0, 10.0], occurrences=-1)

    res = eval_xpath(root, '/fleurInput/cell/bzIntegration/kPointLists/kPointList/kPoint/text()')

    assert res == ['   -0.250000     0.250000     0.000000', '10.0000000000000 10.0000000000000 10.0000000000000']


def test_set_first_text(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_first_text

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    set_first_text(xmltree, schema_dict, 'kPoint', [10.0, 10.0, 10.0])

    res = eval_xpath(root, '/fleurInput/cell/bzIntegration/kPointLists/kPointList/kPoint/text()')

    assert res == ['10.0000000000000 10.0000000000000 10.0000000000000', '    0.250000     0.250000     0.000000']


def test_set_first_text_complex_xpath(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_first_text

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    set_first_text(xmltree,
                   schema_dict,
                   'kPoint', [10.0, 10.0, 10.0],
                   complex_xpath='/fleurInput/cell/bzIntegration/kPointLists/kPointList/kPoint[2]')

    res = eval_xpath(root, '/fleurInput/cell/bzIntegration/kPointLists/kPointList/kPoint/text()')

    assert res == ['   -0.250000     0.250000     0.000000', '10.0000000000000 10.0000000000000 10.0000000000000']


def test_set_first_text_create(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_first_text

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    with pytest.raises(ValueError, match='The tag s has multiple possible paths with the current specification'):
        set_first_text(xmltree, schema_dict, 's', [False, False, False, True])

    with pytest.raises(
            ValueError,
            match=
            "Could not set text on path '/fleurInput/atomSpecies/species/torgueCalculation/greensfElements/s' because atleast one subtag is missing."
    ):
        set_first_text(xmltree, schema_dict, 's', [False, False, False, True], contains={'species', 'torgue'})

    set_first_text(xmltree, schema_dict, 's', [False, False, False, True], contains={'species', 'torgue'}, create=True)

    res = eval_xpath(root, '/fleurInput/atomSpecies/species/torgueCalculation/greensfElements/s/text()')

    assert res == 'F F F T'


def test_add_number_to_attrib(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import add_number_to_attrib

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    add_number_to_attrib(xmltree, schema_dict, 'kmax', 10)

    res = eval_xpath(root, '/fleurInput/calculationSetup/cutoffs/@Kmax')

    assert res == '14.0000000000'


def test_add_number_to_attrib_rel(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import add_number_to_attrib

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    add_number_to_attrib(xmltree, schema_dict, 'kmax', 10, mode='rel')

    res = eval_xpath(root, '/fleurInput/calculationSetup/cutoffs/@Kmax')

    assert res == '40.0000000000'


def test_add_number_to_attrib_specification(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import add_number_to_attrib

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    with pytest.raises(ValueError,
                       match='The attrib radius has multiple possible paths with the current specification.'):
        add_number_to_attrib(xmltree, schema_dict, 'radius', 0.5)

    add_number_to_attrib(xmltree, schema_dict, 'radius', 0.5, not_contains='Group')

    res = eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius')

    assert res == ['2.7000000000', '2.7000000000']


def test_add_number_to_attrib_complex_xpath(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import add_number_to_attrib

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    add_number_to_attrib(xmltree,
                         schema_dict,
                         'radius',
                         0.5,
                         not_contains='Group',
                         complex_xpath="/fleurInput/atomSpecies/species[@name='Pt-1']/mtSphere")

    res = eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius')

    assert res == ['2.20000000', '2.7000000000']


def test_add_number_to_attrib_occurrences(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import add_number_to_attrib

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    add_number_to_attrib(xmltree, schema_dict, 'radius', 0.5, not_contains='Group', occurrences=[0])

    res = eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius')

    assert res == ['2.7000000000', '2.20000000']


def test_add_number_to_first_attrib(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import add_number_to_first_attrib

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    add_number_to_first_attrib(xmltree, schema_dict, 'kmax', 10)

    res = eval_xpath(root, '/fleurInput/calculationSetup/cutoffs/@Kmax')

    assert res == '14.0000000000'


def test_add_number_to_first_attrib_rel(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import add_number_to_first_attrib

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    add_number_to_first_attrib(xmltree, schema_dict, 'kmax', 10, mode='rel')

    res = eval_xpath(root, '/fleurInput/calculationSetup/cutoffs/@Kmax')

    assert res == '40.0000000000'


def test_add_number_to_first_attrib_specification(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import add_number_to_first_attrib

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    with pytest.raises(ValueError,
                       match='The attrib radius has multiple possible paths with the current specification.'):
        add_number_to_first_attrib(xmltree, schema_dict, 'radius', 0.5)

    add_number_to_first_attrib(xmltree, schema_dict, 'radius', 0.5, not_contains='Group')

    res = eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius')

    assert res == ['2.7000000000', '2.20000000']


def test_add_number_to_first_attrib_complex_xpath(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import add_number_to_first_attrib

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    add_number_to_first_attrib(xmltree,
                               schema_dict,
                               'radius',
                               0.5,
                               not_contains='Group',
                               complex_xpath="/fleurInput/atomSpecies/species[@name='Pt-1']/mtSphere")

    res = eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius')

    assert res == ['2.20000000', '2.7000000000']


def test_set_simple_tag(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_simple_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    set_simple_tag(xmltree, schema_dict, 'cutoffs', {'gmax': 34.0, 'gmaxxc': 40.0})

    node = eval_xpath(root, '/fleurInput/calculationSetup/cutoffs')

    assert node.attrib == {'Kmax': '4.00000000', 'Gmax': '34.0000000000', 'GmaxXC': '40.0000000000', 'numbands': '0'}


def test_set_simple_tag_error(load_inpxml):
    from masci_tools.util.xml.xml_setters_names import set_simple_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)

    with pytest.raises(AssertionError, match="Given tag 'species' is not simple"):
        set_simple_tag(xmltree, schema_dict, 'species', {'mtSphere': {'radius': 10}})


def test_set_simple_tag_specification(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_simple_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    with pytest.raises(ValueError, match='The tag lo has multiple possible paths with the current specification.'):
        set_simple_tag(xmltree, schema_dict, 'lo', [{'type': 'TEST', 'n': 12}, {'type': 'TEST', 'n': 15}])

    set_simple_tag(xmltree,
                   schema_dict,
                   'lo', [{
                       'type': 'TEST',
                       'n': 12
                   }, {
                       'type': 'TEST',
                       'n': 15
                   }],
                   contains='species')

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species/lo')

    assert [node.getparent().attrib['name'] for node in nodes] == ['Fe-1', 'Fe-1', 'Pt-1', 'Pt-1']
    assert [node.attrib.items() for node in nodes] == [
        [('type', 'TEST'), ('n', '12')],
        [('type', 'TEST'), ('n', '15')],
        [('type', 'TEST'), ('n', '12')],
        [('type', 'TEST'), ('n', '15')],
    ]


def test_set_simple_tag_create(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_simple_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    with pytest.raises(ValueError, match="Could not create tag 'realAxis' because atleast one subtag is missing."):
        set_simple_tag(xmltree, schema_dict, 'realAxis', {'ne': 100000, 'ellow': -13})

    set_simple_tag(xmltree, schema_dict, 'realAxis', {'ne': 100000, 'ellow': -13}, create_parents=True)

    node = eval_xpath(root, '/fleurInput/calculationSetup/greensFunction/realAxis')

    assert node.attrib == {'ne': '100000', 'ellow': '-13.0000000000'}


def test_set_simple_tag_create_multiple(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_simple_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    with pytest.raises(ValueError,
                       match="Could not create tag 'contourSemicircle' because atleast one subtag is missing."):
        set_simple_tag(xmltree, schema_dict, 'contourSemicircle', [{'n': 12, 'eb': -13}, {'n': 12, 'eb': 55}])

    set_simple_tag(xmltree,
                   schema_dict,
                   'contourSemicircle', [{
                       'n': 12,
                       'eb': -13
                   }, {
                       'n': 12,
                       'eb': 55
                   }],
                   create_parents=True)

    nodes = eval_xpath(root, '/fleurInput/calculationSetup/greensFunction/contourSemicircle')

    assert [node.attrib.items() for node in nodes] == [[('n', '12'), ('eb', '-13.0000000000')],
                                                       [('n', '12'), ('eb', '55.0000000000')]]


def test_set_simple_tag_complex_xpath(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_simple_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    set_simple_tag(xmltree,
                   schema_dict,
                   'lo', [{
                       'type': 'TEST',
                       'n': 12
                   }, {
                       'type': 'TEST',
                       'n': 15
                   }],
                   contains='species',
                   complex_xpath="/fleurInput/atomSpecies/species[@name='Pt-1']")

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species/lo')

    assert [node.getparent().attrib['name'] for node in nodes] == ['Fe-1', 'Fe-1', 'Pt-1', 'Pt-1']
    assert [node.attrib.items() for node in nodes] == [
        [('type', 'SCLO'), ('l', '0'), ('n', '3'), ('eDeriv', '0')],
        [('type', 'SCLO'), ('l', '1'), ('n', '3'), ('eDeriv', '0')],
        [('type', 'TEST'), ('n', '12')],
        [('type', 'TEST'), ('n', '15')],
    ]


def test_set_complex_tag(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_complex_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = {'jspins': 4, 'mtNocoParams': {'l_mtNocoPot': True}, 'qss': [0.0, 1.0, 2.0]}

    set_complex_tag(
        xmltree,
        schema_dict,
        'magnetism',
        changes,
    )

    assert str(eval_xpath(root, '/fleurInput/calculationSetup/magnetism/@jspins')) == '4'
    assert str(eval_xpath(root, '/fleurInput/calculationSetup/magnetism/mtNocoParams/@l_mtNocoPot')) == 'T'
    assert str(eval_xpath(
        root,
        '/fleurInput/calculationSetup/magnetism/qss/text()')) == ' 0.0000000000000  1.0000000000000  2.0000000000000'


def test_set_complex_tag_create_specification(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_complex_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = {'kkintgrcutoff': 'd', 'greensfElements': {'s': [False, True, False, True]}}

    with pytest.raises(ValueError,
                       match='The tag torgueCalculation has multiple possible paths with the current specification.'):
        set_complex_tag(
            xmltree,
            schema_dict,
            'torgueCalculation',
            changes,
        )

    with pytest.raises(
            ValueError,
            match=
            "Could not set attribute 'kkintgrCutoff' on path '/fleurInput/atomSpecies/species/torgueCalculation' because atleast one subtag is missing."
    ):
        set_complex_tag(xmltree, schema_dict, 'torgueCalculation', changes, not_contains='Group')

    set_complex_tag(xmltree, schema_dict, 'torgueCalculation', changes, not_contains='Group', create=True)

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/torgueCalculation/@kkintgrCutoff') == ['d', 'd']
    assert eval_xpath(
        root, '/fleurInput/atomSpecies/species/torgueCalculation/greensfElements/s/text()') == ['F T F T', 'F T F T']


def test_set_complex_tag_complex_xpath(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_complex_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = {'kkintgrcutoff': 'd', 'greensfElements': {'s': [False, True, False, True]}}

    set_complex_tag(xmltree,
                    schema_dict,
                    'torgueCalculation',
                    changes,
                    not_contains='Group',
                    create=True,
                    complex_xpath="/fleurInput/atomSpecies/species[@name='Pt-1']/torgueCalculation")

    node = eval_xpath(root, '/fleurInput/atomSpecies/species/torgueCalculation')

    assert node.getparent().attrib['name'] == 'Pt-1'
    assert eval_xpath(root, '/fleurInput/atomSpecies/species/torgueCalculation/@kkintgrCutoff') == 'd'
    assert eval_xpath(root, '/fleurInput/atomSpecies/species/torgueCalculation/greensfElements/s/text()') == 'F T F T'


def test_set_species(load_inpxml):
    """
    Test of the set_species function (underlying functionality is tested in tests for xml_set_complex_tag)

    Here only the species selection is tested extensively
    """

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_species

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = {
        'mtSphere': {
            'radius': 10.0
        },
        'lo': {
            'type': 'TEST',
            'n': 5,
            'l': 12
        },
        'electronConfig': {
            'stateOccupation': [{
                'state': 'state'
            }, {
                'state': 'state2'
            }]
        }
    }

    set_species(xmltree, schema_dict, 'Fe-1', changes)

    los = eval_xpath(root, '/fleurInput/atomSpecies/species/lo')

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius') == ['10.0000000000', '2.20000000']
    assert eval_xpath(root, '/fleurInput/atomSpecies/species/electronConfig/stateOccupation/@state') == [
        'state', 'state2', '(6s1/2)', '(5d5/2)'
    ]
    assert [lo.attrib for lo in los] == [{
        'type': 'TEST',
        'n': '5',
        'l': '12'
    }, {
        'type': 'SCLO',
        'l': '1',
        'n': '5',
        'eDeriv': '0'
    }]


def test_set_species_label(load_inpxml):
    """
    Test of the set_species function (underlying functionality is tested in tests for xml_set_complex_tag)

    Here only the species selection is tested extensively
    """

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_species_label

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = {
        'mtSphere': {
            'radius': 10.0
        },
        'lo': {
            'type': 'TEST',
            'n': 5,
            'l': 12
        },
        'electronConfig': {
            'stateOccupation': [{
                'state': 'state'
            }, {
                'state': 'state2'
            }]
        }
    }

    set_species_label(xmltree, schema_dict, '                 222', changes)

    los = eval_xpath(root, '/fleurInput/atomSpecies/species/lo')

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius') == ['10.0000000000', '2.20000000']
    assert eval_xpath(root, '/fleurInput/atomSpecies/species/electronConfig/stateOccupation/@state') == [
        'state', 'state2', '(6s1/2)', '(5d5/2)'
    ]
    assert [lo.attrib for lo in los] == [{
        'type': 'TEST',
        'n': '5',
        'l': '12'
    }, {
        'type': 'SCLO',
        'l': '1',
        'n': '5',
        'eDeriv': '0'
    }]


def test_set_species_all(load_inpxml):
    """
    Test of the set_species function (underlying functionality is tested in tests for xml_set_complex_tag)

    Here only the species selection is tested extensively
    """

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_species

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = {
        'mtSphere': {
            'radius': 10.0
        },
        'lo': {
            'type': 'TEST',
            'n': 5,
            'l': 12
        },
        'electronConfig': {
            'stateOccupation': [{
                'state': 'state'
            }, {
                'state': 'state2'
            }]
        }
    }

    set_species(xmltree, schema_dict, 'all', changes)

    los = eval_xpath(root, '/fleurInput/atomSpecies/species/lo')

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius') == ['10.0000000000', '10.0000000000']
    assert eval_xpath(root, '/fleurInput/atomSpecies/species/electronConfig/stateOccupation/@state') == [
        'state', 'state2', 'state', 'state2'
    ]
    assert [lo.attrib for lo in los] == [{'type': 'TEST', 'n': '5', 'l': '12'}, {'type': 'TEST', 'l': '12', 'n': '5'}]


def test_set_species_label_all(load_inpxml):
    """
    Test of the set_species function (underlying functionality is tested in tests for xml_set_complex_tag)

    Here only the species selection is tested extensively
    """

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_species_label

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = {
        'mtSphere': {
            'radius': 10.0
        },
        'lo': {
            'type': 'TEST',
            'n': 5,
            'l': 12
        },
        'electronConfig': {
            'stateOccupation': [{
                'state': 'state'
            }, {
                'state': 'state2'
            }]
        }
    }

    set_species_label(xmltree, schema_dict, 'all', changes)

    los = eval_xpath(root, '/fleurInput/atomSpecies/species/lo')

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius') == ['10.0000000000', '10.0000000000']
    assert eval_xpath(root, '/fleurInput/atomSpecies/species/electronConfig/stateOccupation/@state') == [
        'state', 'state2', 'state', 'state2'
    ]
    assert [lo.attrib for lo in los] == [{'type': 'TEST', 'n': '5', 'l': '12'}, {'type': 'TEST', 'l': '12', 'n': '5'}]


def test_set_species_all_search_string(load_inpxml):
    """
    Test of the set_species function (underlying functionality is tested in tests for xml_set_complex_tag)

    Here only the species selection is tested extensively
    """

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_species

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = {
        'mtSphere': {
            'radius': 10.0
        },
        'lo': {
            'type': 'TEST',
            'n': 5,
            'l': 12
        },
        'electronConfig': {
            'stateOccupation': [{
                'state': 'state'
            }, {
                'state': 'state2'
            }]
        }
    }

    set_species(xmltree, schema_dict, 'all-Pt', changes)

    los = eval_xpath(root, '/fleurInput/atomSpecies/species/lo')

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius') == ['2.20000000', '10.0000000000']
    assert eval_xpath(root, '/fleurInput/atomSpecies/species/electronConfig/stateOccupation/@state') == [
        '(3d3/2)', '(3d5/2)', 'state', 'state2'
    ]
    assert [lo.attrib for lo in los] == [{
        'type': 'SCLO',
        'l': '0',
        'n': '3',
        'eDeriv': '0'
    }, {
        'type': 'SCLO',
        'l': '1',
        'n': '3',
        'eDeriv': '0'
    }, {
        'type': 'TEST',
        'n': '5',
        'l': '12'
    }]


def test_set_atomgroup(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_atomgroup

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = {'force': {'relaxXYZ': 'FFF'}, 'nocoParams': {'beta': 7.0}, 'cFCoeffs': {'potential': True}}

    set_atomgroup(xmltree, schema_dict, changes, position=1)

    assert eval_xpath(root, '/fleurInput/atomGroups/atomGroup/force/@relaxXYZ') == ['FFF', 'TTT']
    assert eval_xpath(root, '/fleurInput/atomGroups/atomGroup/nocoParams/@beta') == ['7.0000000000', '1.570796326']
    assert eval_xpath(root, '/fleurInput/atomGroups/atomGroup/cFCoeffs/@potential') == 'T'


def test_set_atomgroup_species(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_atomgroup

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = {'force': {'relaxXYZ': 'FFF'}, 'nocoParams': {'beta': 7.0}, 'cFCoeffs': {'potential': True}}

    set_atomgroup(xmltree, schema_dict, changes, species='Fe-1')

    assert eval_xpath(root, '/fleurInput/atomGroups/atomGroup/force/@relaxXYZ') == ['FFF', 'TTT']
    assert eval_xpath(root, '/fleurInput/atomGroups/atomGroup/nocoParams/@beta') == ['7.0000000000', '1.570796326']
    assert eval_xpath(root, '/fleurInput/atomGroups/atomGroup/cFCoeffs/@potential') == 'T'


def test_set_atomgroup_all(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_atomgroup

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = {'force': {'relaxXYZ': 'FFF'}, 'nocoParams': {'beta': 7.0}, 'cFCoeffs': {'potential': True}}

    set_atomgroup(xmltree, schema_dict, changes, species='all')

    assert eval_xpath(root, '/fleurInput/atomGroups/atomGroup/force/@relaxXYZ') == ['FFF', 'FFF']
    assert eval_xpath(root, '/fleurInput/atomGroups/atomGroup/nocoParams/@beta') == ['7.0000000000', '7.0000000000']
    assert eval_xpath(root, '/fleurInput/atomGroups/atomGroup/cFCoeffs/@potential') == ['T', 'T']


def test_set_atomgroup_all_position(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_atomgroup

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = {'force': {'relaxXYZ': 'FFF'}, 'nocoParams': {'beta': 7.0}, 'cFCoeffs': {'potential': True}}

    set_atomgroup(xmltree, schema_dict, changes, position='all')

    assert eval_xpath(root, '/fleurInput/atomGroups/atomGroup/force/@relaxXYZ') == ['FFF', 'FFF']
    assert eval_xpath(root, '/fleurInput/atomGroups/atomGroup/nocoParams/@beta') == ['7.0000000000', '7.0000000000']
    assert eval_xpath(root, '/fleurInput/atomGroups/atomGroup/cFCoeffs/@potential') == ['T', 'T']


def test_set_atomgroup_label(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_atomgroup_label

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = {'force': {'relaxXYZ': 'FFF'}, 'nocoParams': {'beta': 7.0}, 'cFCoeffs': {'potential': True}}

    set_atomgroup_label(xmltree, schema_dict, '                 222', changes)

    assert eval_xpath(root, '/fleurInput/atomGroups/atomGroup/force/@relaxXYZ') == ['FFF', 'TTT']
    assert eval_xpath(root, '/fleurInput/atomGroups/atomGroup/nocoParams/@beta') == ['7.0000000000', '1.570796326']
    assert eval_xpath(root, '/fleurInput/atomGroups/atomGroup/cFCoeffs/@potential') == 'T'


def test_set_atomgroup_label_all(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_atomgroup_label

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = {'force': {'relaxXYZ': 'FFF'}, 'nocoParams': {'beta': 7.0}, 'cFCoeffs': {'potential': True}}

    set_atomgroup_label(xmltree, schema_dict, 'all', changes)

    assert eval_xpath(root, '/fleurInput/atomGroups/atomGroup/force/@relaxXYZ') == ['FFF', 'FFF']
    assert eval_xpath(root, '/fleurInput/atomGroups/atomGroup/nocoParams/@beta') == ['7.0000000000', '7.0000000000']
    assert eval_xpath(root, '/fleurInput/atomGroups/atomGroup/cFCoeffs/@potential') == ['T', 'T']


def test_shift_value_species_label(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import shift_value_species_label

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    shift_value_species_label(xmltree, schema_dict, '222', 'lmax', 15)

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/atomicCutoffs/@lmax') == ['25', '10']


def test_shift_value_species_label_rel(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import shift_value_species_label

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    shift_value_species_label(xmltree, schema_dict, '222', 'lmax', 15, mode='rel')

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/atomicCutoffs/@lmax') == ['150', '10']


def test_shift_value_species_label_all(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import shift_value_species_label

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    shift_value_species_label(
        xmltree,
        schema_dict,
        'all',
        'lmax',
        15,
    )

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/atomicCutoffs/@lmax') == ['25', '25']


def test_shift_value_species_label_specification(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import shift_value_species_label

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    with pytest.raises(ValueError, match='The attrib s has multiple possible paths with the current specification.'):
        shift_value_species_label(xmltree, schema_dict, '222', 's', 3)

    shift_value_species_label(xmltree, schema_dict, '222', 's', 3, contains='energyParameters')

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/energyParameters/@s') == ['7', '6']


def test_shift_value_single(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import shift_value

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    shift_value(xmltree, schema_dict, {'itmax': 2})

    assert eval_xpath(root, '/fleurInput/calculationSetup/scfLoop/@itmax') == '3'


def test_shift_value_multiple(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import shift_value

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    shift_value(xmltree, schema_dict, {'itmax': 2, 'kmax': 5.0})

    assert eval_xpath(root, '/fleurInput/calculationSetup/scfLoop/@itmax') == '3'
    assert eval_xpath(root, '/fleurInput/calculationSetup/cutoffs/@Kmax') == '9.0000000000'


def test_shift_value_rel(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import shift_value

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    shift_value(xmltree, schema_dict, {'itmax': 2, 'kmax': 5.0}, mode='rel')

    assert eval_xpath(root, '/fleurInput/calculationSetup/scfLoop/@itmax') == '2'
    assert eval_xpath(root, '/fleurInput/calculationSetup/cutoffs/@Kmax') == '20.0000000000'


def test_shift_value_specification(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import shift_value

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    with pytest.raises(ValueError,
                       match='The attrib spinf has multiple possible paths with the current specification.'):
        shift_value(xmltree, schema_dict, {'itmax': 2, 'kmax': 5.0, 'spinf': 0.5}, mode='rel')

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    shift_value(xmltree,
                schema_dict, {
                    'itmax': 2,
                    'kmax': 5.0,
                    'spinf': 0.5
                },
                mode='rel',
                path_spec={
                    'spinf': {
                        'contains': 'scfLoop'
                    },
                    'KMAX': {
                        'not_contains': 'species'
                    }
                })

    assert eval_xpath(root, '/fleurInput/calculationSetup/scfLoop/@itmax') == '2'
    assert eval_xpath(root, '/fleurInput/calculationSetup/cutoffs/@Kmax') == '20.0000000000'
    assert eval_xpath(root, '/fleurInput/calculationSetup/scfLoop/@spinf') == '1.0000000000'


def test_set_inpchanges_single(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_inpchanges

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    set_inpchanges(xmltree, schema_dict, {'itmax': 20})

    assert eval_xpath(root, '/fleurInput/calculationSetup/scfLoop/@itmax') == '20'


def test_set_inpchanges_single_text(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_inpchanges

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    set_inpchanges(xmltree, schema_dict, {'qss': [10, 10, 10]})

    assert eval_xpath(
        root,
        '/fleurInput/calculationSetup/magnetism/qss/text()') == '10.0000000000000 10.0000000000000 10.0000000000000'


def test_set_inpchanges_multiple(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_inpchanges

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    set_inpchanges(xmltree, schema_dict, {'itmax': 20, 'qss': [10, 10, 10], 'xcFunctional': 'TEST', 'l_linmix': True})

    assert eval_xpath(root, '/fleurInput/calculationSetup/scfLoop/@itmax') == '20'
    assert eval_xpath(
        root,
        '/fleurInput/calculationSetup/magnetism/qss/text()') == '10.0000000000000 10.0000000000000 10.0000000000000'
    assert eval_xpath(root, '/fleurInput/calculationSetup/xcFunctional/@name') == 'TEST'
    assert eval_xpath(root, '/fleurInput/calculationSetup/ldaU/@l_linMix') == 'T'


def test_set_inpchanges_specification(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_names import set_inpchanges

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    with pytest.raises(ValueError,
                       match='The attrib spinf has multiple possible paths with the current specification.'):
        set_inpchanges(xmltree, schema_dict, {
            'itmax': 20,
            'qss': [10, 10, 10],
            'xcFunctional': 'TEST',
            'l_linmix': True,
            'spinf': 10.0
        })

    set_inpchanges(xmltree,
                   schema_dict, {
                       'itmax': 20,
                       'qss': [10, 10, 10],
                       'xcFunctional': 'TEST',
                       'l_linmix': True,
                       'spinf': 10.0
                   },
                   path_spec={
                       'qss': {
                           'contains': 'magnetism'
                       },
                       'SPINF': {
                           'not_contains': 'scfLoop'
                       }
                   })

    assert eval_xpath(root, '/fleurInput/calculationSetup/scfLoop/@itmax') == '20'
    assert eval_xpath(
        root,
        '/fleurInput/calculationSetup/magnetism/qss/text()') == '10.0000000000000 10.0000000000000 10.0000000000000'
    assert eval_xpath(root, '/fleurInput/calculationSetup/xcFunctional/@name') == 'TEST'
    assert eval_xpath(root, '/fleurInput/calculationSetup/ldaU/@l_linMix') == 'T'
    assert eval_xpath(root, '/fleurInput/calculationSetup/ldaU/@spinf') == '10.0000000000'


def test_set_inpchanges_error(load_inpxml):
    from masci_tools.util.xml.xml_setters_names import set_inpchanges

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)

    with pytest.raises(
            ValueError,
            match="You try to set the key:'does_not_exist' to : 'TEST', but the key is unknown to the fleur plug-in"):
        set_inpchanges(xmltree, schema_dict, {
            'itmax': 20,
            'qss': [10, 10, 10],
            'does_not_exist': 'TEST',
            'xcFunctional': 'TEST',
            'l_linmix': True
        })



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module contains a class for organizing and grouping changes to a input file
of fleur in a robust way.

Essentially a low-level version of the FleurinpModifier in aiida_fleur.
"""
from collections import namedtuple
from lxml import etree
import copy
#Enable warnings for missing docstrings
#pylint: enable=missing-function-docstring

ModifierTask = namedtuple('ModifierTask', ['name', 'args', 'kwargs'])


class FleurXMLModifier:
    """
    Class for grouping and organizing changes to a inp.xml file of fleur via the
    xml setting methods in :py:mod:`~masci_tools.util.xml.xml_setters_names` and
    :py:mod:`~masci_tools.util.xml.xml_setters_basic`

    The basic usage is shown below

    .. code-block:: python

        from masci_tools.io.fleurxmlmodifier import FleurXMLModifier

        fmode = FleurXMLModifier()

        #Add changes by calling the methods on this class
        #(names correspond to the setting methods in the xml_setters modules)
        #They are not modifying a input file directly
        #Instead all the tasks are collected and performed in one go

        fmode.set_inpchanges({'Kmax': 4.0}) #Set Kmax to 4.0
        fmode.shift_value({'Gmax': 5.0}) #Add 5 to the current value of Gmax

        #Set the local orbital configuration on all iron atoms to '3s 3p'
        fmode.set_species('all-Fe', {'lo': [{'n':3, 'l': 's', 'type': 'SCLO'},
                                            {'n':3, 'l': 'p', 'type': 'SCLO'}]})

        #To undo the last change call undo
        #fmode.undo()

        #revert_all=True resets all added tasks
        #fmode.undo(revert_all=True)

        #To apply the changes to an input file use the modify_xmlfile method
        new_xmltree = fmode.modify_xmlfile('/path/to/input/file/inp.xml')

    """

    def __init__(self):
        self._tasks = []

    @staticmethod
    def apply_modifications(xmltree, nmmp_lines, modification_tasks, validate_changes=True, extra_funcs=None):
        """
        Applies given modifications to the fleurinp lxml tree.
        It also checks if a new lxml tree is validated against schema.
        Does not rise an error if inp.xml is not validated, simple prints a message about it.

        :param xmltree: a lxml tree to be modified (IS MODIFIED INPLACE)
        :param nmmp_lines: a n_mmp_mat file to be modified (IS MODIFIED INPLACE)
        :param modification_tasks: a list of modification tuples
        :param validate_changes: bool optional (default True), if True after all tasks are performed
                                 both the xmltree and nmmp_lines are checked for consistency
        :param extra_funcs: dict of callables in 'basic', 'schema_dict', 'nmmpmat' subdicts for extra allowed
                            modification functions

        :returns: a modified lxml tree and a modified n_mmp_mat file
        """
        from masci_tools.util.xml.collect_xml_setters import XPATH_SETTERS, SCHEMA_DICT_SETTERS, NMMPMAT_SETTERS
        from masci_tools.util.xml.common_functions import validate_xml, eval_xpath
        from masci_tools.util.xml.xml_setters_nmmpmat import validate_nmmpmat
        from masci_tools.io.parsers.fleur.fleur_schema import InputSchemaDict

        version = eval_xpath(xmltree, '//@fleurInputVersion')
        version = str(version)
        if version is None:
            raise ValueError('Failed to extract inputVersion')

        schema_dict = InputSchemaDict.fromVersion(version)

        xpath_functions = copy.deepcopy(XPATH_SETTERS)
        schema_dict_functions = copy.deepcopy(SCHEMA_DICT_SETTERS)
        nmmpmat_functions = copy.deepcopy(NMMPMAT_SETTERS)

        if extra_funcs is not None:
            xpath_functions.update(extra_funcs.get('basic', {}))
            schema_dict_functions.update(extra_funcs.get('schema_dict', {}))
            nmmpmat_functions.update(extra_funcs.get('nmmpmat', {}))

        for task in modification_tasks:
            if task.name in xpath_functions:
                action = xpath_functions[task.name]
                xmltree = action(xmltree, *task.args, **task.kwargs)

            elif task.name in schema_dict_functions:
                action = schema_dict_functions[task.name]
                xmltree = action(xmltree, schema_dict, *task.args, **task.kwargs)

            elif task.name in nmmpmat_functions:
                action = nmmpmat_functions[task.name]
                nmmp_lines = action(xmltree, nmmp_lines, schema_dict, *task.args, **task.kwargs)

            else:
                raise ValueError(f'Unknown task {task.name}')

        if validate_changes:
            validate_xml(xmltree, schema_dict.xmlschema, error_header='Changes were not valid')

            try:
                validate_nmmpmat(xmltree, nmmp_lines, schema_dict)
            except ValueError as exc:
                msg = f'Changes were not valid (n_mmp_mat file is not compatible): {modification_tasks}'
                raise ValueError(msg) from exc

        return xmltree, nmmp_lines

    def get_avail_actions(self):
        """
        Returns the allowed functions from FleurXMLModifier
        """
        outside_actions = {
            'set_inpchanges': self.set_inpchanges,
            'shift_value': self.shift_value,
            'set_species': self.set_species,
            'set_species_label': self.set_species_label,
            'shift_value_species_label': self.shift_value_species_label,
            'set_atomgroup': self.set_atomgroup,
            'set_atomgroup_label': self.set_atomgroup_label,
            'set_complex_tag': self.set_complex_tag,
            'set_simple_tag': self.set_simple_tag,
            'create_tag': self.create_tag,
            'set_text': self.set_text,
            'set_first_text': self.set_first_text,
            'set_attrib_value': self.set_attrib_value,
            'set_first_attrib_value': self.set_first_attrib_value,
            'add_number_to_attrib': self.add_number_to_attrib,
            'add_number_to_first_attrib': self.add_number_to_first_attrib,
            'xml_create_tag': self.xml_create_tag,
            'xml_replace_tag': self.xml_replace_tag,
            'xml_delete_tag': self.xml_delete_tag,
            'xml_delete_att': self.xml_delete_att,
            'xml_set_attrib_value_no_create': self.xml_set_attrib_value_no_create,
            'xml_set_text_no_create': self.xml_set_text_no_create,
            'set_nmmpmat': self.set_nmmpmat,
            'rotate_nmmpmat': self.rotate_nmmpmat,
        }
        return outside_actions

    def undo(self, revert_all=False):
        """
        Cancels the last change or all of them

        :param revert_all: set True if need to cancel all the changes, False if the last one.
        """
        if revert_all:
            self._tasks = []
        else:
            if self._tasks:
                self._tasks.pop()
                #TODO delete nodes from other nodes
                #del self._tasks[-1]
        return self._tasks

    def changes(self):
        """
        Prints out all changes currently registered on this instance
        """
        from pprint import pprint
        pprint(self._tasks)
        return self._tasks

    def modify_xmlfile(self, original_inpxmlfile, original_nmmp_file=None, validate_changes=True):
        """
        Applies the registered modifications to a given inputfile

        :param original_inpxmlfile: either path to the inp.xml file, opened file handle
                                    or a xml etree to be parsed
        :param original_nmmp_file: path or list of str to a corresponding density matrix
                                   file

        :raises ValueError: if the parsing of the input file

        :returns: a modified xmltree and if existent a modified density matrix file
        """
        if isinstance(original_inpxmlfile, etree._ElementTree):
            original_xmltree = original_inpxmlfile
        else:
            parser = etree.XMLParser(attribute_defaults=True, encoding='utf-8')
            try:
                original_xmltree = etree.parse(original_inpxmlfile, parser)
            except etree.XMLSyntaxError as msg:
                raise ValueError(f'Failed to parse input file: {msg}') from msg

        if original_nmmp_file is not None:
            if isinstance(original_nmmp_file, str):
                with open(original_nmmp_file, mode='r') as n_mmp_file:
                    original_nmmp_lines = n_mmp_file.read().split('\n')
            else:
                original_nmmp_lines = original_nmmp_file
        else:
            original_nmmp_lines = None

        new_xmltree, new_nmmp_lines = self.apply_modifications(original_xmltree,
                                                               original_nmmp_lines,
                                                               self._tasks,
                                                               validate_changes=validate_changes)

        if new_nmmp_lines is None:
            return new_xmltree
        else:
            return new_xmltree, new_nmmp_lines

    def set_inpchanges(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_inpchanges()` to
        the list of tasks that will be done on the xmltree.

        :param change_dict: a dictionary with changes
        :param path_spec: dict, with ggf. necessary further specifications for the path of the attribute

        An example of change_dict::

            change_dict = {'itmax' : 1,
                           'l_noco': True,
                           'ctail': False,
                           'l_ss': True}
        """
        self._tasks.append(ModifierTask('set_inpchanges', args, kwargs))

    def shift_value(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.shift_value()` to
        the list of tasks that will be done on the xmltree.

        :param change_dict: a python dictionary with the keys to shift and the shift values.
        :param mode: 'abs' if change given is absolute, 'rel' if relative
        :param path_spec: dict, with ggf. necessary further specifications for the path of the attribute

        An example of change_dict::

                change_dict = {'itmax' : 1, 'dVac': -0.123}
        """
        self._tasks.append(ModifierTask('shift_value', args, kwargs))

    def set_species(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_species()` to
        the list of tasks that will be done on the xmltree.

        :param species_name: string, name of the specie you want to change
                             Can be name of the species, 'all' or 'all-<string>' (sets species with the string in the species name)
        :param attributedict: a python dict specifying what you want to change.
        :param create: bool, if species does not exist create it and all subtags?

        **attributedict** is a python dictionary containing dictionaries that specify attributes
        to be set inside the certain specie. For example, if one wants to set a MT radius it
        can be done via::

            attributedict = {'mtSphere' : {'radius' : 2.2}}

        Another example::

            'attributedict': {'special': {'socscale': 0.0}}

        that switches SOC terms on a sertain specie. ``mtSphere``, ``atomicCutoffs``,
        ``energyParameters``, ``lo``, ``electronConfig``, ``nocoParams``, ``ldaU`` and
        ``special`` keys are supported. To find possible
        keys of the inner dictionary please refer to the FLEUR documentation flapw.de
        """
        self._tasks.append(ModifierTask('set_species', args, kwargs))

    def set_species_label(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_species_label()` to
        the list of tasks that will be done on the xmltree.

        :param atom_label: string, a label of the atom which specie will be changed. 'all' to change all the species
        :param attributedict: a python dict specifying what you want to change.

        """
        self._tasks.append(ModifierTask('set_species_label', args, kwargs))

    def shift_value_species_label(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.shift_value_species_label()` to
        the list of tasks that will be done on the xmltree.

        :param atom_label: string, a label of the atom which specie will be changed. 'all' if set up all species
        :param attributename: name of the attribute to change
        :param value_given: value to add or to multiply by
        :param mode: 'rel' for multiplication or 'abs' for addition

        Kwargs if the attributename does not correspond to a unique path:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path

        """
        self._tasks.append(ModifierTask('shift_value_species_label', args, kwargs))

    def set_atomgroup(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_atomgroup()` to
        the list of tasks that will be done on the xmltree.

        :param attributedict: a python dict specifying what you want to change.
        :param position: position of an atom group to be changed. If equals to 'all', all species will be changed
        :param species: atom groups, corresponding to the given species will be changed
        :param create: bool, if species does not exist create it and all subtags?

        **attributedict** is a python dictionary containing dictionaries that specify attributes
        to be set inside the certain specie. For example, if one wants to set a beta noco parameter it
        can be done via::

            'attributedict': {'nocoParams': {'beta': val}}

        """
        self._tasks.append(ModifierTask('set_atomgroup', args, kwargs))

    def set_atomgroup_label(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_atomgroup_label()` to
        the list of tasks that will be done on the xmltree.

        :param atom_label: string, a label of the atom which specie will be changed. 'all' to change all the species
        :param attributedict: a python dict specifying what you want to change.
        :param create: bool, if species does not exist create it and all subtags?

        **attributedict** is a python dictionary containing dictionaries that specify attributes
        to be set inside the certain specie. For example, if one wants to set a beta noco parameter it
        can be done via::

            'attributedict': {'nocoParams': {'beta': val}}

        """
        self._tasks.append(ModifierTask('set_atomgroup_label', args, kwargs))

    def create_tag(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.create_tag()` to
        the list of tasks that will be done on the xmltree.

        :param tag_name: str of the tag to create
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param create_parents: bool optional (default False), if True and the given xpath has no results the
                               the parent tags are created recursively
        :param occurrences: int or list of int. Which occurence of the parent nodes to create a tag.
                            By default all nodes are used.

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
        """
        self._tasks.append(ModifierTask('create_tag', args, kwargs))

    def set_complex_tag(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_complex_tag()` to
        the list of tasks that will be done on the xmltree.

        :param tag_name: name of the tag to set
        :param attributedict: Keys in the dictionary correspond to names of tags and the values are the modifications
                              to do on this tag (attributename, subdict with changes to the subtag, ...)
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param create: bool optional (default False), if True and the path, where the complex tag is
                       set does not exist it is created

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path

        """
        self._tasks.append(ModifierTask('set_complex_tag', args, kwargs))

    def set_simple_tag(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_simple_tag()` to
        the list of tasks that will be done on the xmltree.

        :param tag_name: str name of the tag to modify/set
        :param changes: list of dicts or dict with the changes. Elements in list describe multiple tags.
                        Keys in the dictionary correspond to {'attributename': attributevalue}
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param create_parents: bool optional (default False), if True and the path, where the simple tags are
                               set does not exist it is created

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
        """
        self._tasks.append(ModifierTask('set_simple_tag', args, kwargs))

    def set_text(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_text()` to
        the list of tasks that will be done on the xmltree.

        :param tag_name: str name of the tag, where the text should be set
        :param text: value or list of values to set
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param occurrences: int or list of int. Which occurence of the node to set. By default all are set.
        :param create: bool optional (default False), if True the tag is created if is missing

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path

        """
        self._tasks.append(ModifierTask('set_text', args, kwargs))

    def set_first_text(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_first_text()` to
        the list of tasks that will be done on the xmltree.

        :param tag_name: str name of the tag, where the text should be set
        :param text: value or list of values to set
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param create: bool optional (default False), if True the tag is created if is missing

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path

        """
        self._tasks.append(ModifierTask('set_first_text', args, kwargs))

    def set_attrib_value(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_attrib_value()` to
        the list of tasks that will be done on the xmltree.

        :param attributename: the attribute name to set
        :param attribv: value or list of values to set
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param occurrences: int or list of int. Which occurence of the node to set. By default all are set.
        :param create: bool optional (default False), if True the tag is created if is missing

        Kwargs:
            :param tag_name: str, name of the tag where the attribute should be parsed
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param exclude: list of str, here specific types of attributes can be excluded
                            valid values are: settable, settable_contains, other

        """
        self._tasks.append(ModifierTask('set_attrib_value', args, kwargs))

    def set_first_attrib_value(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_first_attrib_value()` to
        the list of tasks that will be done on the xmltree.

        :param attributename: the attribute name to set
        :param attribv: value or list of values to set
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param create: bool optional (default False), if True the tag is created if is missing

        Kwargs:
            :param tag_name: str, name of the tag where the attribute should be parsed
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param exclude: list of str, here specific types of attributes can be excluded
                            valid values are: settable, settable_contains, other

        """
        self._tasks.append(ModifierTask('set_first_attrib_value', args, kwargs))

    def add_number_to_attrib(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.add_number_to_attrib()` to
        the list of tasks that will be done on the xmltree.

        :param attributename: the attribute name to change
        :param add_number: number to add/multiply with the old attribute value
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param mode: str (either `rel` or `abs`).
                     `rel` multiplies the old value with `add_number`
                     `abs` adds the old value and `add_number`
        :param occurrences: int or list of int. Which occurence of the node to set. By default all are set.

        Kwargs:
            :param tag_name: str, name of the tag where the attribute should be parsed
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param exclude: list of str, here specific types of attributes can be excluded
                            valid values are: settable, settable_contains, other

        """
        self._tasks.append(ModifierTask('add_number_to_attrib', args, kwargs))

    def add_number_to_first_attrib(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.add_number_to_first_attrib()` to
        the list of tasks that will be done on the xmltree.

        :param attributename: the attribute name to change
        :param add_number: number to add/multiply with the old attribute value
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param mode: str (either `rel` or `abs`).
                     `rel` multiplies the old value with `add_number`
                     `abs` adds the old value and `add_number`

        Kwargs:
            :param tag_name: str, name of the tag where the attribute should be parsed
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param exclude: list of str, here specific types of attributes can be excluded
                            valid values are: settable, settable_contains, other

        """
        self._tasks.append(ModifierTask('add_number_to_first_attrib', args, kwargs))

    def xml_create_tag(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_basic.xml_create_tag()` to
        the list of tasks that will be done on the xmltree.

        :param xpath: a path where to place a new tag
        :param element: a tag name or etree Element to be created
        :param place_index: defines the place where to put a created tag
        :param tag_order: defines a tag order
        :param occurrences: int or list of int. Which occurence of the parent nodes to create a tag.
                            By default all nodes are used.
        """
        self._tasks.append(ModifierTask('xml_create_tag', args, kwargs))

    def xml_replace_tag(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_basic.xml_replace_tag()` to
        the list of tasks that will be done on the xmltree.

        :param xpath: a path to the tag to be replaced
        :param newelement: a new tag
        """
        self._tasks.append(ModifierTask('xml_replace_tag', args, kwargs))

    def xml_delete_tag(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_basic.xml_delete_tag()` to
        the list of tasks that will be done on the xmltree.

        :param xpath: a path to the tag to be deleted
        """
        self._tasks.append(ModifierTask('xml_delete_tag', args, kwargs))

    def xml_delete_att(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_basic.xml_delete_att()` to
        the list of tasks that will be done on the xmltree.

        :param xpath: a path to the attribute to be deleted
        :param attrib: the name of an attribute
        """
        self._tasks.append(ModifierTask('xml_delete_att', args, kwargs))

    def xml_set_attrib_value_no_create(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_basic.xml_set_attrib_value_no_create()` to
        the list of tasks that will be done on the xmltree.

        :param xpath: a path where to set the attributes
        :param attributename: the attribute name to set
        :param attribv: value or list of values to set (if not str they will be converted with `str(value)`)
        :param occurrences: int or list of int. Which occurence of the node to set. By default all are set.
        """
        self._tasks.append(ModifierTask('xml_set_attrib_value_no_create', args, kwargs))

    def xml_set_text_no_create(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_basic.xml_set_text_no_create()` to
        the list of tasks that will be done on the xmltree.

        :param xpath: a path where to set the attributes
        :param text: value or list of values to set (if not str they will be converted with `str(value)`)
        :param occurrences: int or list of int. Which occurence of the node to set. By default all are set.
        """
        self._tasks.append(ModifierTask('xml_set_text_no_create', args, kwargs))

    def set_nmmpmat(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_nmmpmat.set_nmmpmat()` to
        the list of tasks that will be done on the xmltree.

        :param species_name: string, name of the species you want to change
        :param orbital: integer, orbital quantum number of the LDA+U procedure to be modified
        :param spin: integer, specifies which spin block should be modified
        :param state_occupations: list, sets the diagonal elements of the density matrix and everything
                          else to zero
        :param denmat: matrix, specify the density matrix explicitely
        :param phi: float, optional angle (radian), by which to rotate the density matrix before writing it
        :param theta: float, optional angle (radian), by which to rotate the density matrix before writing it
        """
        self._tasks.append(ModifierTask('set_nmmpmat', args, kwargs))

    def rotate_nmmpmat(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_nmmpmat.rotate_nmmpmat()` to
        the list of tasks that will be done on the xmltree.

        :param species_name: string, name of the species you want to change
        :param orbital: integer, orbital quantum number of the LDA+U procedure to be modified
        :param phi: float, angle (radian), by which to rotate the density matrix
        :param theta: float, angle (radian), by which to rotate the density matrix
        """
        self._tasks.append(ModifierTask('rotate_nmmpmat', args, kwargs))



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module contains the functions necessary to parse mathematical expressions
with predefined constants given in the inp.xml file of Fleur
"""
from __future__ import absolute_import
import numpy as np


def calculate_expression(expression, constants, prevCommand=None, exp_return=False):
    """
    Recursively evaluates the given expression string with the given defined constants

    :param expression: str containing the expression to be parsed
    :param constants: dict with all defined constants (predefined in the Fleur code or defined in the inp.xml)
    :param prevCommand: str, which gives the command before the beginning of the current block
                        if it is given the calculation is stopped, when a command is encountered, which should
                        be exectuted after prevCommand (order of operations)
    :param exp_return: bool, determines whether to return the remaining string of the expression

    :return: float value of the given expression string
    """

    #Map the keywords recognized by fleur to the corresponding numpy function
    functions_dict = {
        'sin': np.sin,
        'cos': np.cos,
        'tan': np.tan,
        'exp': np.exp,
        'log': np.log,
        'abs': np.abs,
        'sqrt': np.sqrt,
        'acos': np.arccos,
        'asin': np.arcsin,
        'atan': np.arctan,
        'cosh': np.cosh,
        'sinh': np.sinh,
        'tanh': np.tanh,
    }

    #Define order of operations
    order_dict = {'+': 10, '-': 10, '*': 100, '/': 100, '%': 100, '**': 1000, '^': 1000}

    stop_loop = False
    loop_count = 0

    if expression is None:
        raise ValueError('Invalid expression: Got None for expression')

    if isinstance(expression, (float, int)):
        return expression

    expression = expression.replace(' ', '')
    value = None
    while not stop_loop and len(expression) != 0:
        loop_count += 1
        firstchar = expression[0]
        if firstchar.isdecimal() or firstchar == '.' or \
           (firstchar in ['+', '-'] and loop_count == 1):
            value, expression = get_first_number(expression)
            stop_loop = False
        elif firstchar.isalpha():
            string, expression = get_first_string(expression)
            if string in functions_dict:
                if expression[0] != '(':
                    raise ValueError('Invalid expression: Expected Bracket after function name')
                function = functions_dict[string]
                function_value, expression = evaluate_bracket(expression, constants)
                #Check conditions for functions
                if string == 'log' and function_value <= 0.0:
                    raise ValueError('Invalid expression: log(x), x<=0')
                elif string == 'sqrt' and function_value < 0.0:
                    raise ValueError('Invalid expression: sqrt(x), x<0')
                elif string == 'asin' and abs(function_value) > 1.0:
                    raise ValueError('Invalid expression: asin(x), |x|>1')
                elif string == 'acos' and abs(function_value) > 1.0:
                    raise ValueError('Invalid expression: acos(x), |x|>1')
                value = function(function_value)
            elif string in constants:
                value = constants[string]
            else:
                raise ValueError(f'Unknown string expression: {string}')
            stop_loop = False
        elif firstchar in ['+', '-', '*', '/', '%', '^']:
            if loop_count == 1:
                raise ValueError(f'Invalid Expression: Found operator {firstchar} in the beginning of expression')
            operator = firstchar
            if expression[1] in ['+', '-', '*', '/', '%', '^']:
                if expression[:2] == '**':
                    operator = expression[:2]
                else:
                    raise ValueError('Invalid Expression: Operator following operator')
            if prevCommand is not None:
                prevOrder = order_dict[prevCommand]
            else:
                prevOrder = 0
            operatorOrder = order_dict[operator]
            if operatorOrder > prevOrder:
                #Evaluate the next block
                block_value, expression = calculate_expression(expression[len(operator):],
                                                               constants,
                                                               prevCommand=operator,
                                                               exp_return=True)
                #Perform the operation
                if operator == '+':
                    value += block_value
                elif operator == '-':
                    value -= block_value
                elif operator == '*':
                    value *= block_value
                elif operator == '/':
                    if abs(block_value) < 1e-12:
                        raise ValueError('Undefined Expression: Division by zero')
                    value *= 1.0 / block_value
                elif operator == '%':
                    value = value % block_value
                elif operator in ['^', '**']:
                    if abs(value) < 1e-12 and abs(block_value) < 1e-12:
                        raise ValueError('Undefined Expression: 0^0')
                    if value < 0.0 and abs(int(block_value) - block_value) > 1e-12:
                        raise ValueError('Undefined Expression: x^y, x<0 and y not integer')
                    elif value < 0.0:
                        block_value = int(block_value)
                    value = value**block_value
            else:
                stop_loop = True
        elif firstchar == '(':
            value, expression = evaluate_bracket(expression, constants)
            stop_loop = False
        else:
            raise ValueError(f'Invalid expression: Found unexpected character {firstchar}')

    if exp_return:
        return value, expression
    else:
        return value


def get_first_number(expression):
    """
    Reads the number in the beginning of the expression string.
    This number can begin with a sign +-, a number or the decimal point

    :param expression: str of the expression

    :return: float value of the number in the beginning and the string of the remaining
             expression
    """
    numberstring = ''
    found_decimal_point = False
    for char in expression:
        if char.isdecimal():
            numberstring += char
        elif char == '.':
            if found_decimal_point:
                raise ValueError('Cannot parse number: Found two decimal points')
            found_decimal_point = True
            numberstring += char
        elif char in ['+', '-']:
            if len(numberstring) == 0:
                numberstring += char
            else:
                break
        else:
            break
    return float(numberstring), expression[len(numberstring):]


def get_first_string(expression):
    """
    Reads the letter string in the beginning of the expression string.

    :param expression: str of the expression

    :return: letter string in the beginning and the string of the remaining
             expression
    """
    found_string = ''
    for char in expression:
        if char.isalpha():
            found_string += char
        else:
            break

    return found_string, expression[len(found_string):]


def evaluate_bracket(expression, constants):
    """
    Evaluates the bracket opened at the start of the expression

    :param expression: expression to be parsed
    :param constants: dict with defined constants

    :return: value of the expression inside the brackets and remaining string of the expression
             after the corresponding closed bracket
    """
    closing_pos = 0
    opened_brackets = 0
    for char in expression:
        if char == '(':
            opened_brackets += 1
        elif char == ')':
            opened_brackets -= 1
        if opened_brackets == 0:
            break
        closing_pos += 1
    if opened_brackets != 0:
        raise ValueError('Invalid Expression: Unbalanced parentheses')

    value = calculate_expression(expression[1:closing_pos], constants)

    return value, expression[closing_pos + 1:]



# -*- coding: utf-8 -*-
"""
Tests for the ParseTasks class
"""
import pytest


def test_default_parse_tasks():
    """
    Test the default parsing tasks fro inconsitencies/typos
    """
    from masci_tools.util.parse_tasks import ParseTasks

    expected_keys = {
        'film_relax_info', 'distances', 'forcetheorem_jij', 'ldau_info', 'bulk_relax_info', 'orbital_magnetic_moments',
        'forcetheorem_mae', 'fleur_modes', 'charges', 'magnetic_distances', 'bandgap', 'forcetheorem_ssdisp',
        'total_energy_contributions', 'ldau_energy_correction', 'general_out_info', 'forces', 'nmmp_distances',
        'total_energy', 'magnetic_moments', 'iteration_number', 'forcetheorem_dmi', 'general_inp_info', 'fermi_energy'
    }

    p = ParseTasks('0.33', validate_defaults=True)

    print(set(p.tasks.keys()))
    assert set(p.tasks.keys()) == expected_keys


def test_find_migration():
    """
    Test the finding of migrations
    """
    from masci_tools.util.parse_tasks import ParseTasks, find_migration

    assert len(find_migration('0.34', '0.34', ParseTasks._migrations)) == 0
    assert len(find_migration('0.34', '0.33', ParseTasks._migrations)) == 1
    assert len(find_migration('0.34', '0.31', ParseTasks._migrations)) == 2
    assert find_migration('0.34', '0.01', ParseTasks._migrations) is None



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module contains custom conversion functions for the outxml_parser, which
cannot be handled by the standard parsing framework
"""
from datetime import date
import numpy as np
from pprint import pprint
from masci_tools.util.constants import HTR_TO_EV
from masci_tools.util.parse_tasks_decorators import conversion_function
from masci_tools.io.common_functions import convert_to_pystd


@conversion_function
def convert_total_energy(out_dict, logger):
    """
    Convert total energy to eV
    """

    total_energy = out_dict.get('energy_hartree', None)

    if total_energy is None:
        if 'energy_hartree' in out_dict:
            if logger is not None:
                logger.warning('convert_total_energy cannot convert None to eV')
            out_dict['energy'] = None
            out_dict['energy_units'] = 'eV'
        return out_dict

    total_energy = total_energy[-1]

    if 'energy' not in out_dict:
        out_dict['energy'] = []
        out_dict['energy_units'] = 'eV'

    if total_energy is not None:
        out_dict['energy'].append(total_energy * HTR_TO_EV)
    else:
        if logger is not None:
            logger.warning('convert_total_energy cannot convert None to eV')
        out_dict['energy'].append(None)

    return out_dict


@conversion_function
def calculate_total_magnetic_moment(out_dict, logger):
    """
    Calculate the the total magnetic moment per cell

    :param out_dict: dict with the already parsed information
    """
    total_charge = out_dict.get('spin_dependent_charge_total', None)

    if total_charge is None:
        if logger is not None:
            logger.warning('calculate_total_magnetic_moment got None')
        return out_dict

    total_charge = total_charge[-1]
    if isinstance(total_charge, list):
        if 'total_magnetic_moment_cell' not in out_dict:
            out_dict['total_magnetic_moment_cell'] = []

        out_dict['total_magnetic_moment_cell'].append(convert_to_pystd(np.abs(total_charge[0] - total_charge[1])))

    return out_dict


@conversion_function
def calculate_walltime(out_dict, logger):
    """
    Calculate the walltime from start and end time

    :param out_dict: dict with the already parsed information
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised
    """

    if out_dict['start_date']['time'] is not None:
        starttimes = out_dict['start_date']['time'].split(':')
    else:
        starttimes = [0, 0, 0]
        msg = 'Starttime was unparsed, inp.xml prob not complete, do not believe the walltime!'
        if logger is not None:
            logger.warning(msg)

    if out_dict['end_date']['time'] is not None:
        endtimes = out_dict['end_date']['time'].split(':')
    else:
        endtimes = [0, 0, 0]
        msg = 'Endtime was unparsed, inp.xml prob not complete, do not believe the walltime!'
        if logger is not None:
            logger.warning(msg)

    if out_dict['start_date']['date'] is not None:
        start_date = out_dict['start_date']['date']
    else:
        start_date = None
        msg = 'Startdate was unparsed, inp.xml prob not complete, do not believe the walltime!'
        if logger is not None:
            logger.warning(msg)

    if out_dict['end_date']['date'] is not None:
        end_date = out_dict['end_date']['date']
    else:
        end_date = None
        msg = 'Enddate was unparsed, inp.xml prob not complete, do not believe the walltime!'
        if logger is not None:
            logger.warning(msg)

    offset = 0
    if start_date is not None and end_date is not None:
        if start_date != end_date:
            date_sl = [int(ent) for ent in start_date.split('/')]
            date_el = [int(ent) for ent in end_date.split('/')]
            date_s = date(*date_sl)
            date_e = date(*date_el)
            diff = date_e - date_s
            offset = diff.days * 86400

    time = offset + (int(endtimes[0]) - int(starttimes[0])) * 60 * 60 + (
        int(endtimes[1]) - int(starttimes[1])) * 60 + int(endtimes[2]) - int(starttimes[2])
    out_dict['walltime'] = time
    out_dict['walltime_units'] = 'seconds'

    return out_dict


@conversion_function
def convert_ldau_definitions(out_dict, logger):
    """
    Convert the parsed information from LDA+U into a more readable dict

    ldau_info has keys for each species with LDA+U ({species_name}/{atom_number})
    and this in turn contains a dict with the LDA+U definition for the given orbital (spdf)

    :param out_dict: dict with the already parsed information
    """
    parsed_ldau = out_dict['ldau_info'].pop('parsed_ldau')
    ldau_species = out_dict['ldau_info'].pop('ldau_species')

    if isinstance(ldau_species['name'], str):
        ldau_species = {key: [val] for key, val in ldau_species.items()}

    if isinstance(parsed_ldau['l'], int):
        parsed_ldau = {key: [val] for key, val in parsed_ldau.items()}

    ldau_definitions = zip(ldau_species['name'], ldau_species['atomic_number'], parsed_ldau['l'])
    for index, ldau_def in enumerate(ldau_definitions):

        species_name, atom_number, orbital = ldau_def

        species_key = f'{species_name}/{atom_number}'
        orbital_key = 'spdf'[orbital]

        if species_key not in out_dict['ldau_info']:
            ldau_dict = out_dict['ldau_info'].get(species_key, {})

        ldau_dict[orbital_key] = {}
        ldau_dict[orbital_key]['u'] = parsed_ldau['u'][index]
        ldau_dict[orbital_key]['j'] = parsed_ldau['j'][index]
        ldau_dict[orbital_key]['unit'] = 'eV'
        if parsed_ldau['l_amf'][index]:
            ldau_dict[orbital_key]['double_counting'] = 'AMF'
        else:
            ldau_dict[orbital_key]['double_counting'] = 'FLL'

        out_dict['ldau_info'][species_key] = ldau_dict

    return out_dict


@conversion_function
def convert_relax_info(out_dict, logger):
    """
    Convert the general relaxation information

    :param out_dict: dict with the already parsed information
    """
    v_1 = out_dict.pop('lat_row1')
    v_2 = out_dict.pop('lat_row2')
    v_3 = out_dict.pop('lat_row3')

    out_dict['relax_brav_vectors'] = [v_1, v_2, v_3]

    out_dict['relax_atom_positions'] = out_dict.pop('atom_positions')
    species = out_dict.pop('position_species')
    species = species['species']
    species_info = out_dict.pop('element_species')
    if isinstance(species_info['name'], str):
        species_info = {key: [val] for key, val in species_info.items()}
    species_info = dict(zip(species_info['name'], species_info['element']))

    out_dict['relax_atomtype_info'] = []
    for specie in species:
        out_dict['relax_atomtype_info'].append((specie, species_info[specie]))

    return out_dict


@conversion_function
def convert_forces(out_dict, logger):
    """
    Convert the parsed forces from a iteration

    :param out_dict: dict with the already parsed information
    """
    parsed_forces = out_dict.pop('parsed_forces')

    if 'force_largest_component' not in out_dict:
        out_dict['force_largest_component'] = []
        out_dict['force_atoms'] = []
        out_dict['abspos_atoms'] = []

    if isinstance(parsed_forces['atom_type'], int):
        parsed_forces = {key: [val] for key, val in parsed_forces.items()}

    largest_force = 0.0
    forces = []
    abspos = []
    for index, atomType in enumerate(parsed_forces['atom_type']):

        force_x = parsed_forces['f_x'][index]
        force_y = parsed_forces['f_y'][index]
        force_z = parsed_forces['f_z'][index]

        x = parsed_forces['x'][index]
        y = parsed_forces['y'][index]
        z = parsed_forces['z'][index]

        forces.append((atomType, [force_x, force_y, force_z]))
        abspos.append((atomType, [x, y, z]))

        if abs(force_x) > largest_force:
            largest_force = abs(force_x)
        if abs(force_y) > largest_force:
            largest_force = abs(force_y)
        if abs(force_z) > largest_force:
            largest_force = abs(force_z)

    out_dict['force_largest_component'].append(largest_force)
    out_dict['force_atoms'].append(forces)
    out_dict['abspos_atoms'].append(abspos)

    return out_dict



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
Tools for the impurity caluclation plugin and its workflows
"""
from sys import version_info
if version_info[0] >= 3:

    def raw_input(msg):
        return eval(eval(input(msg)))


__copyright__ = (u'Copyright (c), 2018, Forschungszentrum Jülich GmbH,' 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.3'
__contributors__ = u'Philipp Rüßmann'


class modify_potential(object):
    """
    Class for old modify potential script, ported from modify_potential script, initially by D. Bauer
    """

    def _check_potstart(self, str1, mode='pot', shape_ver='new'):
        if mode == 'shape':
            if shape_ver == 'new':
                check1 = 'Shape number' in str1
            else:
                check1 = (len(str1) == 11)
        else:
            check1 = 'exc:' in str1
        return check1

    def _read_input(self, filepath):
        #print(filepath)
        with open(filepath) as file:
            data = file.readlines()

        if 'shapefun' in filepath:
            mode = 'shape'
        else:
            mode = 'pot'

        #print(mode, len(data))

        # read file
        index1 = []
        index2 = []
        for i in range(len(data)):
            if self._check_potstart(data[i], mode=mode):
                index1.append(i)
                if len(index1) > 1:
                    index2.append(i - 1)
        index2.append(i)

        # read shapefun if old style is used
        if mode == 'shape' and len(index1) < 1:
            index1 = []
            index2 = []
            for i in range(len(data)):
                if self._check_potstart(data[i], mode=mode, shape_ver='old'):
                    index1.append(i)
                if len(index1) > 1:
                    index2.append(i - 1)
                index2.append(i)
        """
        print(index1)
        print(index2)

        print('Potential file read')
        print('found %i potentials in file'%len(index1))
        print('')
        """

        return index1, index2, data

    def shapefun_from_scoef(self, scoefpath, shapefun_path, atom2shapes, shapefun_new):
        """
        Read shapefun and create impurity shapefun using scoef info and shapes array

        :param scoefpath: absolute path to scoef file
        :param shapefun_path: absolute path to input shapefun file
        :param shapes: shapes array for mapping between atom index and shapefunction index
        :param shapefun_new: absolute path to output shapefun file to which the new shapefunction will be written
        """
        index1, index2, data = self._read_input(shapefun_path)

        order = list(range(len(index1)))

        natomtemp = int(open(scoefpath).readlines()[0])
        filedata = open(scoefpath).readlines()[1:natomtemp + 1]
        listnew = []
        for line in filedata:
            if len(line.split()) > 1:
                listnew.append(atom2shapes[int(line.split()[3]) - 1] - 1)
        order = listnew

        datanew = []
        for i in range(len(order)):
            for ii in range(index1[order[i]], index2[order[i]] + 1):
                datanew.append(data[ii])

        # add header to shapefun_new
        tmp = datanew
        datanew = []
        datanew.append('   %i\n' % (len(order)))
        datanew.append('  1.000000000000E+00\n')
        datanew += tmp
        with open(shapefun_new, 'w') as f:
            f.writelines(datanew)

    def neworder_potential(self, potfile_in, potfile_out, neworder, potfile_2=None, replace_from_pot2=None):
        """
        Read potential file and new potential using a list describing the order of the new potential.
        If a second potential is given as input together with an index list, then the corresponding of
        the output potential are overwritten with positions from the second input potential.

        :param potfile_in: absolute path to input potential
        :type potfile_in: str
        :param potfile_out: absolute path to output potential
        :type potfile_out: str
        :param neworder: list after which output potential is constructed from input potential
        :type neworder: list
        :param potfile_2: optional, absolute path to second potential file if
            positions in new list of potentials shall be replaced by positions of
            second potential, requires *replace_from_pot* to be given as well
        :type potfile_2: str
        :param replace_from_pot: optional, list containing tuples of (position
            in newlist that is to be replaced, position in pot2 with which position
            is replaced)
        :type replace_from_pot: list

        :usage:
            1. modify_potential().neworder_potential(<path_to_input_pot>, <path_to_output_pot>, [])
        """
        from numpy import array, shape

        index1, index2, data = self._read_input(potfile_in)

        if potfile_2 is not None:
            index12, index22, data2 = self._read_input(potfile_2)
            # check if also replace_from_pot2 is given correctly
            if replace_from_pot2 is None:
                raise ValueError('replace_from_pot2 not given')
            else:
                replace_from_pot2 = array(replace_from_pot2)
                if shape(replace_from_pot2)[1] != 2:
                    raise ValueError('replace_from_pot2 needs to be a 2D array!')
        else:
            if replace_from_pot2 is not None:
                raise ValueError('replace_from_pot2 given but potfile_2 not given')

        # set order in which potential file is written
        # ensure that numbers are integers:
        order = [int(i) for i in neworder]

        datanew = []
        for i in range(len(order)):
            # check if new position is replaced with position from old pot
            if replace_from_pot2 is not None and i in replace_from_pot2[:, 0]:
                replace_index = replace_from_pot2[replace_from_pot2[:, 0] == i][0][1]
                for ii in range(index12[replace_index], index22[replace_index] + 1):
                    datanew.append(data2[ii])
            else:  # otherwise take new potntial according to input list
                for ii in range(index1[order[i]], index2[order[i]] + 1):
                    datanew.append(data[ii])

        # write out new potential
        with open(potfile_out, 'w') as f:
            f.writelines(datanew)



# -*- coding: utf-8 -*-
"""
Tests for the load functions in io_fleurxml
"""
from lxml import etree
import os
import pytest

FILE_PATH = os.path.dirname(os.path.abspath(__file__))


def test_load_inpxml():
    from masci_tools.io.io_fleurxml import load_inpxml

    TEST_INPXML_PATH = os.path.join(FILE_PATH, 'files/fleur/Max-R5/FePt_film_SSFT_LO/files/inp2.xml')

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)

    assert xmltree is not None
    assert schema_dict['inp_version'] == '0.34'

    #Parse before
    parser = etree.XMLParser(attribute_defaults=True, encoding='utf-8')
    xmltree = etree.parse(TEST_INPXML_PATH, parser)
    xmltree, schema_dict = load_inpxml(xmltree)

    assert xmltree is not None
    assert schema_dict['inp_version'] == '0.34'

    #Pass file handle
    with open(TEST_INPXML_PATH, 'r') as inpfile:
        xmltree, schema_dict = load_inpxml(inpfile)

    assert xmltree is not None
    assert schema_dict['inp_version'] == '0.34'


def test_load_outxml():
    from masci_tools.io.io_fleurxml import load_outxml

    TEST_OUTXML_PATH = os.path.join(FILE_PATH, 'files/fleur/Max-R5/SiLOXML/files/out.xml')

    xmltree, schema_dict = load_outxml(TEST_OUTXML_PATH)

    assert xmltree is not None
    assert schema_dict['out_version'] == '0.34'
    assert schema_dict['inp_version'] == '0.34'

    #Parse before
    parser = etree.XMLParser(attribute_defaults=True, encoding='utf-8')
    xmltree = etree.parse(TEST_OUTXML_PATH, parser)
    xmltree, schema_dict = load_outxml(xmltree)

    assert xmltree is not None
    assert schema_dict['out_version'] == '0.34'
    assert schema_dict['inp_version'] == '0.34'

    #Pass file handle
    with open(TEST_OUTXML_PATH, 'r') as inpfile:
        xmltree, schema_dict = load_outxml(inpfile)

    assert xmltree is not None
    assert schema_dict['out_version'] == '0.34'
    assert schema_dict['inp_version'] == '0.34'


def test_loadoutxml_mixedversions():
    from masci_tools.io.io_fleurxml import load_outxml

    TEST_OUTXML_PATH = os.path.join(FILE_PATH, 'files/fleur/output_mixed_versions.xml')

    xmltree, schema_dict = load_outxml(TEST_OUTXML_PATH)

    assert xmltree is not None
    assert schema_dict['out_version'] == '0.34'
    assert schema_dict['inp_version'] == '0.33'


def test_loadoutxml_max50():
    from masci_tools.io.io_fleurxml import load_outxml

    TEST_OUTXML_PATH = os.path.join(FILE_PATH, 'files/fleur/old_versions/Max5_0_test_out.xml')

    with pytest.warns(UserWarning):
        xmltree, schema_dict = load_outxml(TEST_OUTXML_PATH)

    assert xmltree is not None
    assert schema_dict['out_version'] == '0.33'
    assert schema_dict['inp_version'] == '0.33'


def test_loadoutxml_max40():
    from masci_tools.io.io_fleurxml import load_outxml

    TEST_OUTXML_PATH = os.path.join(FILE_PATH, 'files/fleur/old_versions/Max4_test_out.xml')

    with pytest.warns(UserWarning):
        xmltree, schema_dict = load_outxml(TEST_OUTXML_PATH)

    assert xmltree is not None
    assert schema_dict['out_version'] == '0.31'
    assert schema_dict['inp_version'] == '0.31'


def test_loadoutxml_max31():
    from masci_tools.io.io_fleurxml import load_outxml

    TEST_OUTXML_PATH = os.path.join(FILE_PATH, 'files/fleur/old_versions/Max3_1_test_out.xml')

    with pytest.warns(UserWarning):
        xmltree, schema_dict = load_outxml(TEST_OUTXML_PATH)

    assert xmltree is not None
    assert schema_dict['out_version'] == '0.30'
    assert schema_dict['inp_version'] == '0.30'


def test_loadoutxml_premax31():
    from masci_tools.io.io_fleurxml import load_outxml

    TEST_OUTXML_PATH = os.path.join(FILE_PATH, 'files/fleur/old_versions/Max3_0_test_out.xml')

    with pytest.warns(UserWarning):
        xmltree, schema_dict = load_outxml(TEST_OUTXML_PATH)

    assert xmltree is not None
    assert schema_dict['out_version'] == '0.29'
    assert schema_dict['inp_version'] == '0.29'
